
with Asis.Ada_Environments;
with Asis.Implementation;
with Asis.Compilation_Units;

with Ada.Directories;
with Tiny.Texts;

package body ASIS_Vendors.GNAT is

   package Implementation    renames Asis.Implementation;
   package Ada_Environments  renames Asis.Ada_Environments;
   package Compilation_Units renames Asis.Compilation_Units;

   -- ------------------------------------------------------------------------

   Opened_Context_Count : Natural := 0;

   -- ------------------------------------------------------------------------

   procedure Begin_With_Context
     (Context : out Asis.Context; Name : Wide_String)
   is
      Parameters : constant Wide_String := "-CA -FT -SA -T.";
   begin

      if Opened_Context_Count = 0 then
         Implementation.Initialize;
      end if;

      Ada_Environments.Associate
        (The_Context => Context,
         Name        => Name,
         Parameters  => Parameters);

      Ada_Environments.Open (Context);

      Opened_Context_Count := Opened_Context_Count + 1;

   end Begin_With_Context;

   -- ------------------------------------------------------------------------

   procedure End_With_Context
     (Context : in out Asis.Context)
   is
   begin
      if Opened_Context_Count = 0 then
         raise Program_Error;
      end if;

      Ada_Environments.Close (Context);
      Ada_Environments.Dissociate (Context);

      Opened_Context_Count := Opened_Context_Count - 1;

      if Opened_Context_Count = 0 then
         Implementation.Finalize;
      end if;
   end End_With_Context;

   -- ------------------------------------------------------------------------

   function Unit_Name_From_File_Name
     (Context   : in Asis.Context;
      File_Name : in Wide_String)
      return Wide_String
   is
      Units : constant Asis.Compilation_Unit_List :=
        Compilation_Units.Compilation_Units (Context);
      -- Neither Compilation_Units.Compilation_Units,
      -- Compilation_Units.Library_Unit_Declarations or
      -- Compilation_Units.Compilation_Unit_Bodies seems to be able to
      -- list specifications with ASIS-for-GNAT : unit declarations are never
      -- retrieved, so we cannot find an *.ads file, only an *.adb file.
   begin
      Validity_Constraints :
      begin
         if not (File_Name'Length /= 0) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         package Directories renames Ada.Directories;

         Standard_Specification_Extension : constant Wide_String := "ads";
         Standard_Body_Extension : constant Wide_String := "adb";

         -- We have to apply mutiple To_String <-> To_Wide_String convertion,
         -- because the text type used for ASIS file-names is not the same
         -- as the text type used for Ada file-names.

         Full_Name : constant Wide_String :=
           Tiny.Texts.To_Wide_String
             (Directories.Full_Name
                  (Tiny.Texts.To_String (File_Name)));
         -- ASIS for GNAT seems to always use absolute paths (Full_Name).
         -- Although we later create Standard_Full_Name_For_Specification and
         -- Standard_Full_Name_For_Body, this Full_Name is still used,
         -- because extension for unit may vary in a way this application
         -- cannot determines, and using the originally provided one, give us
         -- more chance successfully find a match.
         Extension : constant Wide_String :=
           Tiny.Texts.To_Wide_String
             (Directories.Extension
                  (Tiny.Texts.To_String (Full_Name)));
         -- To remove the current extension and then append two other ones to
         -- two new paths (file name)
         Full_Name_Without_Extension : constant Wide_String :=
           Full_Name (Full_Name'First .. Full_Name'Last - Extension'Length);
         -- To later append two extensions to this
         Standard_Full_Name_For_Specification : constant Wide_String :=
           Full_Name_Without_Extension & Standard_Specification_Extension;
         -- Remember, we will still test the Full_Name any way.
         Standard_Full_Name_For_Body : constant Wide_String :=
           Full_Name_Without_Extension & Standard_Body_Extension;
         -- Remember, we will still test the Full_Name any way.
      begin
         for I in Units'Range loop
            Test_Unit_File_Name :
            declare
               Unit : constant Asis.Compilation_Unit := Units (I);
            begin
               if Compilation_Units.Text_Name (Unit) = Full_Name then
                  return Compilation_Units.Unit_Full_Name (Unit);
               end if;
               if
                 (Compilation_Units.Text_Name (Unit)) =
                 (Standard_Full_Name_For_Specification)
               then
                  return Compilation_Units.Unit_Full_Name (Unit);
               end if;
               if
                 (Compilation_Units.Text_Name (Unit)) =
                 (Standard_Full_Name_For_Body)
               then
                  return Compilation_Units.Unit_Full_Name (Unit);
               end if;
            end Test_Unit_File_Name;
         end loop;
         return Not_A_Unit_Name;
      exception
         when Directories.Name_Error =>
            return Not_A_Unit_Name;
      end Common_Case;
   end Unit_Name_From_File_Name;

end ASIS_Vendors.GNAT;
