
--! Do not modify this pattern, use a copy with a meaningful name instead.
--! See comments marked with "--!" to know where you should update this
--! implementation pattern.

with Asis.Ada_Environments;
with Asis.Implementation;
with Asis.Compilation_Units;

with Ada.Directories;
with Tiny.Texts;

package body ASIS_Vendors.Typical is

   package Implementation    renames Asis.Implementation;
   package Ada_Environments  renames Asis.Ada_Environments;
   package Compilation_Units renames Asis.Compilation_Units;

   -- ------------------------------------------------------------------------

   Opened_Context_Count : Natural := 0;

   -- ------------------------------------------------------------------------

   procedure Begin_With_Context
     (Context : out Asis.Context; Name : Wide_String)
   is
      Parameters : constant Wide_String := "";
      --! Update this constant to match your compiler/Asis-implementation.
      --! See README.txt to learn more about the requirements which these
      --! parameters must fullfill.
   begin

      if Opened_Context_Count = 0 then
         Implementation.Initialize;
      end if;

      Ada_Environments.Associate
        (The_Context => Context,
         Name        => Name,
         Parameters  => Parameters);

      Ada_Environments.Open (Context);

      Opened_Context_Count := Opened_Context_Count + 1;

   end Begin_With_Context;

   -- ------------------------------------------------------------------------

   procedure End_With_Context
     (Context : in out Asis.Context)
   is
   begin
      if Opened_Context_Count = 0 then
         raise Program_Error;
      end if;

      Ada_Environments.Close (Context);
      Ada_Environments.Dissociate (Context);

      Opened_Context_Count := Opened_Context_Count - 1;

      if Opened_Context_Count = 0 then
         Implementation.Finalize;
      end if;
   end End_With_Context;

   -- ------------------------------------------------------------------------

   function Unit_Name_From_File_Name
     (Context   : in Asis.Context;
      File_Name : in Wide_String)
      return Wide_String
   is
      Units : constant Asis.Compilation_Unit_List :=
        Compilation_Units.Compilation_Units (Context);
      -- Neither Compilation_Units.Compilation_Units,
      -- Compilation_Units.Library_Unit_Declarations or
      -- Compilation_Units.Compilation_Unit_Bodies seems to be able to
      -- list specifications with Asis-for-GNAT : unit declarations are never
      -- retrieved, so we cannot find an *.ads file, only an *.adb file.
   begin
      Validity_Constraints :
      begin
         if not (File_Name'Length /= 0) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         package Directories renames Ada.Directories;

         -- We have to apply mutiple To_String <-> To_Wide_String convertion,
         -- because the text type used for Asis file-names is not the same
         -- as the text type used for Ada file-names.

         Full_Name : constant Wide_String :=
           Tiny.Texts.To_Wide_String
             (Directories.Full_Name
                  (Tiny.Texts.To_String (File_Name)));
         -- An Asis implementation may use relative of absolute path, or
         -- not file at all. We will compare to the full name path and the
         -- relative or absolute path given as an argument to this method.
         --! Update this to match the specification of your compiler/Asis-
         --! implementation. If your compiler does not make use of files,
         --! please send a message to one of the contact given in the
         --! Contacts section of README.txt. If the compiler associated to
         --! the Asis implementation makes use of files, and if it use
         --! separated files for specification and implementation or
         --! even use any additional files, then this method should be$
         --! able to retrieve a unit name from a file name referring to either
         --! an implemtation file, a specification file or any other file
         --! which is part of the source of a given unit.
      begin
         for I in Units'Range loop
            Test_Unit_File_Name :
            declare
               Unit : constant Asis.Compilation_Unit := Units (I);
            begin
               if Compilation_Units.Text_Name (Unit) = File_Name then
                  return Compilation_Units.Unit_Full_Name (Unit);
               end if;
               if Compilation_Units.Text_Name (Unit) = Full_Name then
                  return Compilation_Units.Unit_Full_Name (Unit);
               end if;
            end Test_Unit_File_Name;
         end loop;
         return Not_A_Unit_Name;
      exception
         when Directories.Name_Error =>
            return Not_A_Unit_Name;
      end Common_Case;
   end Unit_Name_From_File_Name;

end ASIS_Vendors.Typical;
