
with Shell_UI.Application_Description;
with Shell_UI.Command_Line_Options;
with Shell_UI.Standard_Command_Line_Options;
with Shell_UI.Help;

with ASISEyes;

with Tiny.Texts;
with Ada.Text_IO;
with Asis.Exceptions;

procedure ASISEyes_Application
is
   package Exceptions renames Asis.Exceptions;
   package IO renames Ada.Text_IO;

   ASIS_Error_Indicator : constant String :=
     ("An ASIS error occured");

   ASIS_Error_Handling_Suggestion : constant String :=
     ("Please, ensure the ASIS environement is up to date");

begin

   Application_Description :
   declare
      package Application_Description renames
        Shell_UI.Application_Description;

      Application_Name : constant String :=
        "ASISEyes";

      Application_Purpose : constant String :=
        "Show the ASIS view of an Ada source text";

      Main_Copyrigth : constant String :=
        "Y.D. 2010 - See README.txt for license details";

      ASIS_Component_Name : constant String :=
        "Ada Semantic Interface Specification (ASIS)";

      ASIS_Copyright : constant String :=
        "ISO/IEC 1999";

      Binary_Dependencies_Component_Name : constant String :=
        "Dependencies";

      Binary_Dependencies_Copyright : constant String :=
        "Depending on your compiler/ASIS vendor";
   begin
      Application_Description.Set_Name
        (Application_Name);

      Application_Description.Set_Purpose_Text
        (Application_Purpose);

      Application_Description.Set_Main_Copyright_Notice
        (Main_Copyrigth);

      Application_Description.Add_Third_Party_Copyright_Notice
        (Component_Name   => ASIS_Component_Name,
         Copyright_Notice => ASIS_Copyright);

      Application_Description.Add_Third_Party_Copyright_Notice
        (Component_Name   => Binary_Dependencies_Component_Name,
         Copyright_Notice => Binary_Dependencies_Copyright);

      Application_Description.Set_Version
        (Major  => 1,
         Minor  => 0,
         Status => Application_Description.Release);
   end Application_Description;

   Command_Line_Interpretation :
   declare
      function To_Wide_String
        (S : String)
      return Wide_String
      renames Tiny.Texts.To_Wide_String;

      package Options renames Shell_UI.Command_Line_Options;
      package Standard_Options renames Shell_UI.Standard_Command_Line_Options;
      package Help renames Shell_UI.Help;

      Unit_Name_Option_Short_Name : constant Character := 'u';
      Unit_Name_Long_Name         : constant String := "unit-name";
      Is_Unit_Name_A_Flag         : constant Boolean := False;
      Unit_Name_Option_Type_Name  : constant String := "name";
      Unit_Name_Option_Help       : constant String :=
        ("Fully qualified name of an Ada compilation unit, either body, " &
         "specification or program. As with any Ada name identifier, this " &
         "is a case insensitive name.");

      File_Name_Option_Short_Name : constant Character := 'f';
      File_Name_Long_Name         : constant String := "file-name";
      Is_File_Name_A_Flag         : constant Boolean := False;
      File_Name_Option_Type_Name  : constant String := "path";
      File_Name_Option_Help       : constant String :=
        ("Fully qualified absolute file name, with extension, of an " &
         "Ada compilation unit, either body, specification or program. " &
         "Note : depending on your compiler vendor, the file name may be " &
         "required to stands for a compilable unit, ex. a body unit. " &
         "Depending on your host operating-system, the file name may also " &
         "be case sensitive or case insensitive. Be aware of these points " &
         "or else ASISEyes will not be able to find your unit.");

      Configuration_Option_Short_Name : constant Character := 'c';
      Configuration_Long_Name         : constant String := "asis-config";
      Is_Configuration_A_Flag         : constant Boolean := True;
      Configuration_Option_Type_Name  : constant String := "boolean";
      Configuration_Option_Help       : constant String :=
        ("Display ASIS configuration informations, such as implementor, " &
         "version and the likes, then exit.");

      Message_When_Nor_File_Name_Nor_Unit_Name_Was_Provided
        : constant String :=
          ("At least one of the unit name or the file name must be " &
           "provided.");

      Unit_Name_Option : constant Options.Option_Access :=
        Options.Declare_Option
          (Short_Name => Unit_Name_Option_Short_Name,
           Long_Name  => Unit_Name_Long_Name,
           Is_A_Flag  => Is_Unit_Name_A_Flag,
           Type_Name  => Unit_Name_Option_Type_Name,
           Help_Text  => Unit_Name_Option_Help);

      File_Name_Option : constant Options.Option_Access :=
        Options.Declare_Option
          (Short_Name => File_Name_Option_Short_Name,
           Long_Name  => File_Name_Long_Name,
           Is_A_Flag  => Is_File_Name_A_Flag,
           Type_Name  => File_Name_Option_Type_Name,
           Help_Text  => File_Name_Option_Help);

      Configuration_Option : constant Options.Option_Access :=
        Options.Declare_Option
          (Short_Name => Configuration_Option_Short_Name,
           Long_Name  => Configuration_Long_Name,
           Is_A_Flag  => Is_Configuration_A_Flag,
           Type_Name  => Configuration_Option_Type_Name,
           Help_Text  => Configuration_Option_Help);
   begin
      Standard_Options.Add_Standard_Options;
      Options.Parse_Command_Line;
      Standard_Options.First_Handle_Standard_Options;

      if Options.Provided (Unit_Name_Option) then
         -- Do not check for Configuration_Option, as this work is included
         -- in the work done for ASISEyes_Using_Unit_Name
         ASISEyes.ASISEyes_Using_Unit_Name
           (To_Wide_String (Options.Value (Unit_Name_Option)));
      elsif Options.Provided (File_Name_Option) then
         -- Do not check for Configuration_Option, as this work is included
         -- in the work done for ASISEyes_Using_File_Name
         ASISEyes.ASISEyes_Using_File_Name
           (To_Wide_String (Options.Value (File_Name_Option)));
      elsif Options.Provided (Configuration_Option) then
         ASISEyes.Put_ASIS_Configuration;
      else
         Help.Display_Usage
           (Message => Message_When_Nor_File_Name_Nor_Unit_Name_Was_Provided);
      end if;

   end Command_Line_Interpretation;

exception
   when Exceptions.ASIS_Failed =>
      IO.Put_Line (ASIS_Error_Indicator);
      IO.Put_Line (ASIS_Error_Handling_Suggestion);
   when Shell_UI.Standard_Command_Line_Options.Quit_Application =>
      null;
      -- A normal request to qui the application : no notification.
   when Shell_UI.Command_Line_Options.Command_Line_Error =>
      null;
      -- The Command_Line_Options package has already displayed a
      -- message : do not add more.
end ASISEyes_Application;
