
with Ada.Text_IO;

package body IDE_Vendors.GPS is

   package IO renames Ada.Text_IO;

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. End of paraggraph
   -- 2. Raw messages
   -- 3. Message associated to a file/text location
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. End of paraggraph
   -- ------------------------------------------------------------------------

   procedure End_Of_Paragraph
   is
   begin
      IO.New_Line;
   end End_Of_Paragraph;

   -- ========================================================================
   -- 2. Raw messages
   -- ------------------------------------------------------------------------
   -- 2.1 From string abstraction
   -- 2.2 From string access
   -- 2.3 From string
   -- ------------------------------------------------------------------------

   -- ------------------------------------------------------------------------
   -- 2.1 From string abstraction
   -- ------------------------------------------------------------------------

   procedure Log_Message (Log_Text : in out Text_To_Write_Class)
   is
   begin
      Write (Log_Text);
      IO.New_Line;
   end Log_Message;

   -- ------------------------------------------------------------------------
   -- 2.2 From string access
   -- ------------------------------------------------------------------------

   procedure Log_Message (Log_Text : access constant String)
   is
   begin
      -- For consistency, reuse the implementation which is based on
      -- a string access. This implementation is thus a
      -- wrapper.
      Log_Message (Log_Text.all);
   end Log_Message;

   -- ------------------------------------------------------------------------
   -- 2.3 From string
   -- ------------------------------------------------------------------------

   procedure Log_Message (Log_Text : in String)
   is
   begin
      Validity_Constraints :
      begin
         if not (Log_Text'Length > 0) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         -- For consistency, reuse the implementation which is based on
         -- an abstract text parameter. This implementation is thus a
         -- wrapper.

         type Text_To_Write_Type is new IDE_Vendors.Text_To_Write_Type
         with null record;

         procedure Write (Text_To_Write : Text_To_Write_Type)
         is
         begin
            pragma Unreferenced (Text_To_Write);
            -- Use the outer Log_Text parameter only
            IO.Put (Log_Text);
         end Write;

         Text_To_Write : Text_To_Write_Type;
      begin
         Log_Message (Text_To_Write);
      end Common_Case;
   end Log_Message;

   -- ========================================================================
   -- 3. Message associated to a file/text location
   -- ------------------------------------------------------------------------
   -- 2.1 Format parameters
   -- 2.2 From string abstraction
   -- 2.3 From string access
   -- 2.4 From string
   -- ------------------------------------------------------------------------

   -- ------------------------------------------------------------------------
   -- 2.1 Format parameters
   -- ------------------------------------------------------------------------

   File_Name_To_Line_Number_Separator     : constant String := ":";
   Line_Number_To_Column_Number_Separator : constant String := ":";
   Column_Number_To_Message_Separator     : constant String := ": ";

   -- Notice the leading space in Column_Number_To_Message_Separator

   -- ------------------------------------------------------------------------
   -- 2.2 From string abstraction
   -- ------------------------------------------------------------------------

   procedure Log_Message_With_File_Position
     (File_Name     : access constant String;
      Line_Number   : in Positive;
      Column_Number : in Positive;
      Log_Text      : in out Text_To_Write_Class)
   is
   begin
      Validity_Constraints :
      begin
         if not (File_Name'Length > 0) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         Line_Number_Image : constant String :=
           Positive'Image (Line_Number);

         Column_Number_Image : constant String :=
           Positive'Image (Column_Number);

         Space             : constant Character := ' ';

         function Less_Heading_Space (S : in String) return String
         --|Requires: S'Length > 1
         --|Requires: S (S'First) = ' '
         is
         begin
            Validity_Constraints :
            begin
               if not (S'Length > 1) then
                  raise Program_Error;
               end if;

               if S (S'First) /= Space then
                  raise Program_Error;
               end if;
            end Validity_Constraints;

            return S (S'First + 1 .. S'Last);
         end Less_Heading_Space;

      begin
         IO.Put (File_Name.all);
         IO.Put (File_Name_To_Line_Number_Separator);
         IO.Put (Less_Heading_Space (Line_Number_Image));
         IO.Put (Line_Number_To_Column_Number_Separator);
         IO.Put (Less_Heading_Space (Column_Number_Image));
         IO.Put (Column_Number_To_Message_Separator);
         Write (Log_Text);
         IO.New_Line;
      end Common_Case;
   end Log_Message_With_File_Position;

   -- ------------------------------------------------------------------------
   -- 2.3 From string access
   -- ------------------------------------------------------------------------

   procedure Log_Message_With_File_Position
     (File_Name     : access constant String;
      Line_Number   : in Positive;
      Column_Number : in Positive;
      Log_Text      : access constant String)
   is
   begin
      -- For consistency, reuse the implementation which is based on
      -- an abstract text parameter. This implementation is thus a
      -- wrapper.

      Log_Message_With_File_Position
        (File_Name     => File_Name,
         Line_Number   => Line_Number,
         Column_Number => Column_Number,
         Log_Text      => Log_Text.all);

   end Log_Message_With_File_Position;

   -- ------------------------------------------------------------------------
   -- 2.4 From string
   -- ------------------------------------------------------------------------

   procedure Log_Message_With_File_Position
     (File_Name     : access constant String;
      Line_Number   : in Positive;
      Column_Number : in Positive;
      Log_Text      : in String)
   is
   begin
      Validity_Constraints :
      begin
         if not (File_Name'Length > 0) then
            raise Program_Error;
         end if;
         if not (Log_Text'Length > 0) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         -- For consistency, reuse the implementation which is based on
         -- an abstract text parameter. This implementation is thus a
         -- wrapper.

         type Text_To_Write_Type is new IDE_Vendors.Text_To_Write_Type
         with null record;

         procedure Write (Text_To_Write : Text_To_Write_Type)
         is
         begin
            pragma Unreferenced (Text_To_Write);
            -- Use the outer Log_Text parameter only
            IO.Put (Log_Text);
         end Write;

         Text_To_Write : Text_To_Write_Type;
      begin
         Log_Message_With_File_Position
           (File_Name     => File_Name,
            Line_Number   => Line_Number,
            Column_Number => Column_Number,
            Log_Text      => Text_To_Write);
      end Common_Case;
   end Log_Message_With_File_Position;

end IDE_Vendors.GPS;
