
with Tiny.Containers;

with Tiny.Texts;
with Ada.Command_Line;
with Shell_UI.Help;
with Shell_UI.Formated_Strings;

pragma Elaborate_All (Tiny.Containers);

package body Shell_UI.Command_Line_Options is

   package Texts renames Tiny.Texts;
   package Formating renames Shell_UI.Formated_Strings;

   -- ========================================================================
   -- Text table
   -- ------------------------------------------------------------------------

   Unknown_Parameter_Option : constant String :=
     ("The parameter ""$1"" is an unknown option of this application.");

   Option_Already_Assigned_A_Value : constant String :=
     ("The option ""$1"" was already previoulsy assigned a value.");

   Option_Requires_To_Be_Assigned_A_Value : constant String :=
     ("The option ""$1"" requires to be assigned a value.");

   Flag_Option_Cannot_Be_Assigned_A_Value : constant String :=
     ("The option ""$1"" is a flag option, and thus cannot be assigned " &
      "a value.");

   Option_Cannot_Be_Assigned_Multiple_Values : constant String :=
     ("The option ""$1"" cannot be assigned multiple values.");

   Option_Name_Expected : constant String :=
     ("An option name was expected, found ""$1"" instead.");

   Short_Option_Name_Too_Long : constant String :=
     ("The parameter ""$1"" does not match a valid option name format. It " &
      "has a short option name prefix, but it has a too much long name.");

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Command line option type
   -- 2. Command line options list and iteration
   -- 3. Parsing command line in accordance to command line options
   -- 4. Adding/Creating command line options
   -- 5. Accessing command line options
   -- 6. Text formating concerns
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. Command line option type
   -- ------------------------------------------------------------------------
   -- 1.1 Command line option type
   -- 1.2 Command line options list
   -- ------------------------------------------------------------------------

   -- ------------------------------------------------------------------------
   -- 1.1 Command line option type
   -- ------------------------------------------------------------------------

   type Option_Type
     (Long_Name_Length : Natural;
      Type_Name_Length : Natural;
      Help_Text_Length : Natural)
   is
      record
         Short_Name  : Character;
         Long_Name   : String (1 .. Long_Name_Length);
         Is_A_Flag   : Boolean;
         Type_Name   : String (1 .. Type_Name_Length);
         Help_Text   : String (1 .. Help_Text_Length);
         Value       : access String := null;
         Flagged     : Boolean := False;
      end record;

   -- ------------------------------------------------------------------------

   function Short_Name
     (Option : Option_Access)
      return Character
   is
   begin
      return Option.Short_Name;
   end Short_Name;

   -- ------------------------------------------------------------------------

   function Long_Name
     (Option : Option_Access)
      return String
   is
   begin
      return Option.Long_Name;
   end Long_Name;

   -- ------------------------------------------------------------------------

   function Name
     (Option : Option_Access)
      return String
   is
   begin
      if Option.Long_Name /= Option_Null_Long_Name then
         if Option.Short_Name /= Option_Null_Short_Name then
            return
              (Option.Long_Name &
               Formating.Alternative_Operator &
               Option.Short_Name);
         else
            return Option.Long_Name;
         end if;
      else
         return Texts.To_String (Option.Short_Name);
      end if;
   end Name;

   -- ------------------------------------------------------------------------

   function Is_A_Flag
     (Option : Option_Access)
      return Boolean
   is
   begin
      return Option.Is_A_Flag;
   end Is_A_Flag;

   -- ------------------------------------------------------------------------

   function Type_Name
     (Option : Option_Access)
      return String
   is
   begin
      return Option.Type_Name;
   end Type_Name;

   -- ------------------------------------------------------------------------

   function Help_Text
     (Option : Option_Access)
      return String
   is
   begin
      return Option.Help_Text;
   end Help_Text;

   -- ========================================================================
   -- 2. Command line options list and iteration
   -- ------------------------------------------------------------------------

   package Command_Line_Option_Lists is
     new Tiny.Containers.Vectors
       (Element_Type => Option_Access);

   package Options_Lists renames Command_Line_Option_Lists;

   -- ------------------------------------------------------------------------

   Options : Options_Lists.Vector_Access := Options_Lists.Null_Vector;

   -- ------------------------------------------------------------------------

   procedure Read_Each_Options
     (Accessor : in out Option_Accessor_Class)
   -- Provides access to the options list on the client side
   is
      Client_Accessor : Option_Accessor_Class renames Accessor;

      type Accessor_Wrapper_Type is new Options_Lists.Accessor_Type
      with null record;

      procedure Do_Access
        (Accessor : in out Accessor_Wrapper_Type;
         Element  : in Option_Access) is
      begin
         pragma Unreferenced (Accessor);
         Do_Access (Accessor => Client_Accessor, Option => Element);
      end Do_Access;

      Accessor_Wrapper : Accessor_Wrapper_Type;

   begin
      Options_Lists.For_Each_In
        (Accessor   => Accessor_Wrapper,
         Vector     => Options);
   end Read_Each_Options;

   -- ------------------------------------------------------------------------

   function Options_List_Is_Empty return Boolean
   is
   begin
      return Options_Lists.Is_Null (Options);
   end Options_List_Is_Empty;

   -- ========================================================================
   -- 3. Parsing command line in accordance to command line options
   -- ------------------------------------------------------------------------
   -- 3.1 Status and error
   -- 3.2 Parsing
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 3.1 Status and error
   -- ------------------------------------------------------------------------

   Command_Line_Parsed_Status : Boolean := False;

   -- ------------------------------------------------------------------------

   function Command_Line_Parsed return Boolean
   is
   begin
      return Command_Line_Parsed_Status;
   end Command_Line_Parsed;

   -- ------------------------------------------------------------------------

   procedure Raise_Command_Line_Error (Message : in String)
   is
   begin
      Shell_UI.Help.Display_Usage (Message => Message);
      Ada.Command_Line.Set_Exit_Status (Ada.Command_Line.Failure);
      raise Command_Line_Error;
   end Raise_Command_Line_Error;

   -- ========================================================================
   -- 3.2 Parsing
   -- ------------------------------------------------------------------------
   -- 3.2.1 Option name kind
   -- 3.2.2 Option name from parameter text
   -- 3.2.3 Option access from name
   -- 3.2.4 Current option
   -- 3.2.5 Check option states
   -- 3.2.6 Main

   procedure Parse_Command_Line
   is

      -- -------------------------------------
      -- 3.2.1 Option name kind
      -- -------------------------------------

      type Option_Name_Kinds is
        (Not_An_Option_Name,
         An_Option_Short_Name,
         An_Option_Long_Name);

      -- -------------------------------------

      function Option_Name_Kind (Text : String) return Option_Name_Kinds
      is
      begin
         if Text'Length = 0 then
            raise Program_Error;
         end if;

         if Text'Length = 2 then
            if Text (Text'First) = Dash then
               if Text (Positive'Succ (Text'First)) = Dash then
                  return Not_An_Option_Name;
               end if;
               return An_Option_Short_Name;
            end if;
         end if;

         if Text'Length > 2 then
            if Text (Text'First) = Dash then
               if Text (Positive'Succ (Text'First)) = Dash then
                  return An_Option_Long_Name;
               else
                  -- We have something of the kind -xxxx, which
                  -- is not a valid option name format.
                  Raise_Command_Line_Error
                    (Texts.Variable_Message
                       (Pattern => Short_Option_Name_Too_Long,
                        Value   => Text));
               end if;
            end if;
         end if;

         return Not_An_Option_Name;

      end Option_Name_Kind;

      -- -------------------------------------
      -- 3.2.2 Option name from parameter text
      -- -------------------------------------

      function Option_Long_Name (Text : String) return String
      is
         First : Positive := Text'First;
         Last  : constant Positive := Text'Last;
      begin
         First := Positive'Succ (First);
         -- Skeep the first Dash

         First := Positive'Succ (First);
         -- Skeep the second Dash

         return Text (First .. Last);
      end Option_Long_Name;

      -- -------------------------------------

      function Option_Short_Name (Text : String) return Character
      is
      begin
         return Text (Positive'Succ (Text'First));
      end Option_Short_Name;

      -- -------------------------------------
      -- 3.2.3 Option access from name
      -- -------------------------------------

      procedure Get_Option_For_Long_Name
        (Long_Name : in String;
         Found     : out Boolean;
         Option    : out Option_Access)
      --|Ensures: Found is defined
      --|Ensures: (Found = False) => (Option = old Option)
      is
         type Long_Name_Test_Type is new Options_Lists.Test_Type
         with null record;

         function Checked
           (Test    : Long_Name_Test_Type;
            Element : Option_Access)
            return Boolean
         is
         begin
            pragma Unreferenced (Test);
            return Element.Long_Name = Long_Name;
         end Checked;

         Long_Name_Test : Long_Name_Test_Type;
      begin
         Options_Lists.Find_In
           (Vector    => Options,
            Test      => Long_Name_Test,
            Found     => Found,
            Element   => Option);
      end Get_Option_For_Long_Name;

      -- -------------------------------------

      procedure Get_Option_For_Short_Name
        (Short_Name : in Character;
         Found      : out Boolean;
         Option     : out Option_Access)
      --|Ensures: Found is defined
      --|Ensures: (Found = False) => (Option = old Option)
      is
         type Short_Name_Test_Type is new Options_Lists.Test_Type
         with null record;

         function Checked
           (Test    : Short_Name_Test_Type;
            Element : Option_Access)
            return Boolean
         is
         begin
            pragma Unreferenced (Test);
            return Element.Short_Name = Short_Name;
         end Checked;

         Short_Name_Test : Short_Name_Test_Type;
      begin
         Options_Lists.Find_In
           (Vector    => Options,
            Test      => Short_Name_Test,
            Found     => Found,
            Element   => Option);
      end Get_Option_For_Short_Name;

      -- -------------------------------------
      -- 3.2.4 Current option
      -- -------------------------------------

      Expect_An_Option_Name : Boolean := True;
      -- We must always start with an option name.

      Option_To_Assign_Value_To : Option_Access := null;

      -- -------------------------------------
      -- 3.2.5 Check option states
      -- -------------------------------------

      procedure Check_Option_Was_Found
        (Founded        : Boolean;
         Parameter_Text : String)
      is
      begin
         if not Founded then
            Raise_Command_Line_Error
              (Texts.Variable_Message
                 (Pattern => Unknown_Parameter_Option,
                  Value   => Parameter_Text));
         end if;
      end Check_Option_Was_Found;

      -- -------------------------------------

      procedure Check_Option_Is_Not_Already_Assigned_A_Value
        (Option         : Option_Access;
         Parameter_Text : String)
      is
      begin
         if Option.Value /= null then
            Raise_Command_Line_Error
              (Texts.Variable_Message
                 (Pattern => Option_Already_Assigned_A_Value,
                  Value   => Name (Option)));
         end if;
      end Check_Option_Is_Not_Already_Assigned_A_Value;

      -- -------------------------------------

      procedure Check_Previous_Option_Has_Its_Required_Value
      is
      begin
         if Option_To_Assign_Value_To = null then
            return;
         end if;
         if Option_To_Assign_Value_To.Is_A_Flag then
            return;
         end if;
         if Option_To_Assign_Value_To.Value = null then
            Raise_Command_Line_Error
              (Texts.Variable_Message
                 (Pattern => Option_Requires_To_Be_Assigned_A_Value,
                  Value   => Name (Option_To_Assign_Value_To)));
         end if;
      end Check_Previous_Option_Has_Its_Required_Value;

      -- -------------------------------------
      -- 3.2.6 Main
      -- -------------------------------------

   begin
      Command_Line_Parsed_Status := True;
      -- Mark command line option as parsed when we start to parse it,
      -- to allow usage help to be displayed whenever an error occured
      -- while the command line is parsed.

      if Ada.Command_Line.Argument_Count = 0 then
         return;
      end if;

      Iterate_Parameters :
      for I in 1 .. Ada.Command_Line.Argument_Count loop
         Interpret_Parameter_Text :
         declare
            Parameter_Text : constant String := Ada.Command_Line.Argument (I);
         begin
            case Option_Name_Kind (Parameter_Text) is
               when Not_An_Option_Name =>
                  if Expect_An_Option_Name then
                     Raise_Command_Line_Error
                       (Texts.Variable_Message
                          (Pattern => Option_Name_Expected,
                           Value   => Parameter_Text));
                  end if;
                  if Option_To_Assign_Value_To.Is_A_Flag then
                     Raise_Command_Line_Error
                       (Texts.Variable_Message
                          (Pattern => Flag_Option_Cannot_Be_Assigned_A_Value,
                           Value   => Name (Option_To_Assign_Value_To)));
                  end if;
                  if Option_To_Assign_Value_To.Value /= null then
                     Raise_Command_Line_Error
                       (Texts.Variable_Message
                          (Pattern =>
                             Option_Cannot_Be_Assigned_Multiple_Values,
                           Value   =>
                             Name (Option_To_Assign_Value_To)));
                  end if;
                  Option_To_Assign_Value_To.Value :=
                    new String'(Ada.Command_Line.Argument (I));
               when An_Option_Short_Name =>
                  Check_Previous_Option_Has_Its_Required_Value;
                  declare
                     Found      : Boolean;
                     Short_Name : constant Character :=
                       Option_Short_Name (Parameter_Text);
                  begin
                     Get_Option_For_Short_Name
                       (Short_Name => Short_Name,
                        Found      => Found,
                        Option     => Option_To_Assign_Value_To);
                     Check_Option_Was_Found
                       (Founded        => Found,
                        Parameter_Text => Parameter_Text);
                     Check_Option_Is_Not_Already_Assigned_A_Value
                       (Option         => Option_To_Assign_Value_To,
                        Parameter_Text => Parameter_Text);
                     if Option_To_Assign_Value_To.Is_A_Flag then
                        Option_To_Assign_Value_To.Flagged := True;
                     end if;
                  end;
               when An_Option_Long_Name =>
                  Check_Previous_Option_Has_Its_Required_Value;
                  declare
                     Found     : Boolean;
                     Long_Name : constant String :=
                       Option_Long_Name (Parameter_Text);
                  begin
                     Get_Option_For_Long_Name
                       (Long_Name  => Long_Name,
                        Found      => Found,
                        Option     => Option_To_Assign_Value_To);
                     Check_Option_Was_Found
                       (Founded        => Found,
                        Parameter_Text => Parameter_Text);
                     Check_Option_Is_Not_Already_Assigned_A_Value
                       (Option         => Option_To_Assign_Value_To,
                        Parameter_Text => Parameter_Text);
                     if Option_To_Assign_Value_To.Is_A_Flag then
                        Option_To_Assign_Value_To.Flagged := True;
                     end if;
                  end;
            end case;
         end Interpret_Parameter_Text;
         Expect_An_Option_Name := False;
      end loop Iterate_Parameters;

      Check_Previous_Option_Has_Its_Required_Value;
      -- In the loop, this check is done for a the current argument, on
      -- the previous argument. If we want the check to be done on the
      -- last argument, which has no following argument, we must do it
      -- after the loop.

   end Parse_Command_Line;


   -- ========================================================================
   -- 4. Adding/Creating command line options
   -- ------------------------------------------------------------------------
   -- 4.1 Command line options short and long names
   -- 4.2 Adding/Creating options
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 4.1 Command line options short and long names
   -- ------------------------------------------------------------------------

   function Option_Short_Name_Already_Used
     (Short_Name : Character)
      return Boolean
   is

      type Short_Option_Name_Test_Type is new Options_Lists.Test_Type
      with null record;

      -- -------------------------------------

      function Checked
        (Test    : Short_Option_Name_Test_Type;
         Element : Option_Access)
         return Boolean is
      begin
         pragma Unreferenced (Test);
         return Element.Short_Name = Short_Name;
      end Checked;

      -- -------------------------------------

      Test : Short_Option_Name_Test_Type;
   begin
      Special_Cases :
      begin
         if Short_Name = Option_Null_Short_Name then
            return False;
         end if;
      end Special_Cases;

      return Options_Lists.Exist_In (Vector => Options, Test => Test);
   end Option_Short_Name_Already_Used;

   -- ------------------------------------------------------------------------

   function Option_Long_Name_Already_Used
     (Long_Name : String)
      return Boolean
   is

      type Long_Option_Name_Test_Type is new Options_Lists.Test_Type
      with null record;

      -- -------------------------------------

      function Checked
        (Test    : Long_Option_Name_Test_Type;
         Element : Option_Access)
         return Boolean is
      begin
         pragma Unreferenced (Test);
         return Element.Long_Name = Long_Name;
      end Checked;

      -- -------------------------------------

      Test : Long_Option_Name_Test_Type;
   begin
      Special_Cases :
      begin
         if Long_Name = Option_Null_Long_Name then
            return False;
         end if;
      end Special_Cases;

      return Options_Lists.Exist_In (Vector => Options, Test => Test);
   end Option_Long_Name_Already_Used;

   -- ========================================================================
   -- 4.2 Adding/Creating options
   -- ------------------------------------------------------------------------

   function Declare_Option
     (Short_Name : in Character;
      Long_Name  : in String;
      Is_A_Flag  : in Boolean;
      Type_Name  : in String;
      Help_Text  : in String)
      return Option_Access
   is
   begin

      -- -------------------------------------

      Check_Validity_Constraints :
      begin
         if Command_Line_Parsed then
            raise Program_Error;
         end if;

         if
           (Short_Name = Option_Null_Short_Name) and
           (Long_Name = Option_Null_Long_Name)
         then
            raise Program_Error;
         end if;

         if not Texts.Is_Graphic (Short_Name) then
            if not (Short_Name = Option_Null_Short_Name) then
               raise Program_Error;
            end if;
         end if;

         if not Texts.Is_Graphic (Long_Name) then
            if not (Long_Name = Option_Null_Long_Name) then
               raise Program_Error;
            end if;
         end if;

         if Option_Short_Name_Already_Used (Short_Name) then
            raise Program_Error;
         end if;

         if Option_Long_Name_Already_Used (Long_Name) then
            raise Program_Error;
         end if;

         if Long_Name'Length < Minimum_Option_Long_Name_Length then
            if Long_Name = Option_Null_Long_Name then
               null;
               -- Accept
            else
               raise Program_Error;
            end if;
         end if;

         if Help_Text'Length < Minimum_Option_Help_Text_Length then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;

      -- -------------------------------------

      Do_Valid_Operation :
      declare
         Option : constant Option_Access :=
           new Option_Type
             (Long_Name_Length => Long_Name'Length,
              Type_Name_Length => Type_Name'Length,
              Help_Text_Length => Help_Text'Length);
      begin

         Initialize_Option :
         begin
            Option.Short_Name := Short_Name;

            Option.Long_Name (1 .. Option.Long_Name_Length) :=
              Long_Name (Long_Name'First .. Long_Name'Last);

            Option.Is_A_Flag := Is_A_Flag;

            Option.Type_Name (1 .. Option.Type_Name_Length) :=
              Type_Name (Type_Name'First .. Type_Name'Last);

            Option.Help_Text (1 .. Option.Help_Text_Length) :=
              Help_Text (Help_Text'First .. Help_Text'Last);

            Option.Value := null;
         end Initialize_Option;

         Append_Option :
         begin
            Options_Lists.Append_To
              (Vector    => Options,
               Element   => Option);
         end Append_Option;

         return Option;

      end Do_Valid_Operation;

   end Declare_Option;

   -- ========================================================================
   -- 5. Accessing command line options
   -- ------------------------------------------------------------------------

   function Provided
     (Option : Option_Access)
      return Boolean
   is
   begin

      Validity_Constraints :
      begin
         if not Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Special_Cases :
      begin
         if Option = Null_Option then
            return False;
         end if;
      end Special_Cases;

      return (Option.Value /= null) or (Option.Flagged);

   end Provided;

   -- ------------------------------------------------------------------------

   function Value
     (Option : Option_Access)
      return String
   is
   begin
      Validity_Constraints :
      begin
         if not Command_Line_Parsed then
            raise Program_Error;
         end if;

         if not Provided (Option) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      return Option.Value.all;

   end Value;

   -- ------------------------------------------------------------------------

   function Flagged
     (Option : Option_Access)
      return Boolean
   is
   begin
      Validity_Constraints :
      begin
         if not Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      if Option = Null_Option then
         return False;
      else
         return Option.Flagged;
      end if;

   end Flagged;

   -- ========================================================================
   -- 6. Text formating concerns
   -- ------------------------------------------------------------------------

   function Option_Short_Name_Text (Option_Name : Character) return String
   is
   begin
      return Option_Short_Name_Prefix & Option_Name;
   end Option_Short_Name_Text;

   -- ------------------------------------------------------------------------

   function Option_Long_Name_Text (Option_Name : String) return String
   is
   begin
      if not (Option_Name'Length > 1) then
         raise Program_Error;
      end if;

      return Option_Long_Name_Prefix & Option_Name;
   end Option_Long_Name_Text;

end Shell_UI.Command_Line_Options;
