
-- Defines command line options by their short name and long names, with
-- type hint and help text. Provide methods to parse the command line
-- using these command line options definitions. See also
-- Standard_Command_Line_Options.
--
-- When an error encountered in the command line, the package displays the
-- command line option help with a message which explains the error, then
-- it raise Command_Line_Error. Clients of this package should handle this
-- exception so that the client application do nothing, as the package
-- as already displayed a message (unless it has more to add), and terminates
-- silently (without displaying an exception trace back).
--
-- Command line option help is provided by the Help package and is not
-- part of this one.

package Shell_UI.Command_Line_Options is

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Command line option type
   -- 2. Command line options list and iteration
   -- 3. Parsing command line in accordance to command line options
   -- 4. Adding/Creating command line options
   -- 5. Accessing command line options
   -- 6. Text formating concerns
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. Command line option type
   -- ------------------------------------------------------------------------

   type Option_Access is private;

   Null_Option : constant Option_Access;

   function Short_Name
     (Option : Option_Access)
      return Character;

   function Long_Name
     (Option : Option_Access)
      return String;

   function Name
     (Option : Option_Access)
      return String;

   function Is_A_Flag
     (Option : Option_Access)
      return Boolean;

   function Type_Name
     (Option : Option_Access)
      return String;

   function Help_Text
     (Option : Option_Access)
      return String;

   -- ========================================================================
   -- 2. Command line options list and iteration
   -- ------------------------------------------------------------------------

   type Option_Accessor_Type is abstract tagged private;
   -- Question : as the implementation obviously makes use of some kind of
   -- list to handle options list, why not to simply provide the definition
   -- of this list ?
   -- Answer : there may be multiple choices for such a list, and client of
   -- this package are better to not depends on this choice. Also, the
   -- purpose of this package is not to provide containers. So finally, the
   -- list is left implementation defined, and this simple accessor is
   -- provided. Any way, clients of this package are not supposed to do
   -- anything else on the option list than iteratively reading all of
   -- the options.

   procedure Do_Access
     (Accessor : in out Option_Accessor_Type;
      Option   : in Option_Access)
     is abstract;
   -- To be use with Read_Each_Option
   -- This method should be named after the verb Access, but this is
   -- an Ada reserved word, so it is named Do_Access.

   subtype Option_Accessor_Class is
     Option_Accessor_Type'Class;

   procedure Read_Each_Options
     (Accessor : in out Option_Accessor_Class);

   function Options_List_Is_Empty return Boolean;
   -- This is the only function dealing with options list cardinality.
   -- There is no function which give the number of arguments, either
   -- declared or provided arguments, just because this would not comes
   -- with any meaningfully added value. Moreover, there is no way to
   -- legitimately define a Command_Line_Options_Count_Type. Just the list
   -- is meaningfull, not any cardinality.

   -- ========================================================================
   -- 3. Parsing command line in accordance to command line options
   -- ------------------------------------------------------------------------

   function Command_Line_Parsed return Boolean;
   -- Initialy False

   Command_Line_Error : exception;
   -- When this exception is raised, the command line option package has
   -- already displayed usage informations. The main program either simply do
   -- nothing when catching this option or display further more informations
   -- if there are some to show.

   procedure Raise_Command_Line_Error (Message : in String);
   -- Raise Command_Line_Error after, but takes care to display the help
   -- before.

   procedure Parse_Command_Line;
   --|Ensures: Command_Line_Parsed
   --
   -- The client is responsible to check for required options.
   --
   -- All declared options are supposed to be optional.
   --
   -- No Required property is defined on command line options in this package,
   -- as this property may depends on complex combinations of options.
   --
   -- This package will raise Command_Line_Error via Raise_Command_Line_Error
   -- if an unknown option is encountered, if the command line is
   -- hill-formed, if a given option appears multiple times in the command
   -- line, if a flag option is gived a literal value, if a option which is
   -- not a flag option is not given a literal value.
   --
   -- If a required command line option is missing, the client may raise
   -- Command_Line_Error using Raise_Command_Line_Error.

   -- ========================================================================
   -- 4. Adding/Creating command line options
   -- ------------------------------------------------------------------------
   -- 4.1 Command line options short and long names
   -- 4.2 Adding/Creating options
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 4.1 Command line options short and long names
   -- ------------------------------------------------------------------------

   Option_Null_Short_Name : constant Character := ' ';
   Option_Null_Long_Name  : constant String := "";

   Minimum_Option_Long_Name_Length : constant := 2;
   -- At least two characters long, to not conflict with option short names

   Minimum_Option_Help_Text_Length : constant := 12;
   -- At least two or three words is a minimum. A too much high value would
   -- drive to unmeaningful wording where just few words is needed.

   function Option_Short_Name_Already_Used
     (Short_Name : Character)
      return Boolean;
   -- Initialy False for any Name
   --|Ensures: Short_Name = Option_Null_Short_Name -> Result = False

   function Option_Long_Name_Already_Used
     (Long_Name : String)
      return Boolean;
   -- Initialy False for any Name
   --|Ensures: Long_Name = Option_Null_Long_Name -> Result = False

   -- ========================================================================
   -- 4.2 Adding/Creating options
   -- ------------------------------------------------------------------------

   function Declare_Option
     (Short_Name : in Character;
      Long_Name  : in String;
      Is_A_Flag  : in Boolean;
      Type_Name  : in String;
      Help_Text  : in String)
      return Option_Access;
   --|Requires: not Command_Line_Parsed
   --|Requires:
   --|   (Short_Name /= Option_Null_Short_Name) or
   --|   (Long_Name /= Option_Null_Long_Name)
   --|Requires:
   --|   (Is_Graphic (Short_Name)) or else
   --|   (Short_Name = Option_Null_Short_Name)
   --|Requires:
   --|   (Is_Graphic (Long_Name)) or else
   --|   (Long_Name = Option_Null_Long_Name)
   --|Requires: not Short_Option_Name_Already_Used (Short_Name)
   --|Requires: not Long_Option_Name_Already_Used (Long_Name)
   --|Requires:
   --|   (Long_Name'Length >= Minimum_Option_Long_Name_Length) or else
   --|   (Long_Name = Option_Null_Long_Name)
   --|Requires:
   --|   Option_Help'Length >= Minimum_Option_Help_Text_Length
   --|Ensures:
   --|   (Short_Name /= Option_Null_Short_Name) <->
   --|   (Short_Name_Already_Used (Short_Name))
   --|Ensures:
   --|   (Long_Name /= Option_Null_Long_Name) <->
   --|   (Long_Name_Already_Used (Long_Name))
   --|Ensures: Result /= Null_Option
   --|Ensures: Short_Name (Result) = Short_Name
   --|Ensures: Long_Name (Result) = Long_Name
   --|Ensures: Is_A_Flag (Result) = Is_A_Flag
   --|Ensures: Type_Name (Result) = Type_Name
   --|Ensures: Help_Text (Result) = Help_Text

   -- ========================================================================
   -- 5. Accessing command line options
   -- ------------------------------------------------------------------------

   function Provided
     (Option : Option_Access)
      return Boolean;
   --|Requires: Command_Line_Parsed
   --|Ensures: Option = Null_Option -> Result = False

   function Value
     (Option : Option_Access)
      return String;
   --|Requires: Command_Line_Parsed and then Provided (Option)

   function Flagged
     (Option : Option_Access)
      return Boolean;
   --|Requires: Command_Line_Parsed
   --|Ensures: Provided (Option) -> Result = True
   -- Returns True for an option which is provided on the command line
   -- (and was not given a value, otherwise Parse_Command_Line would have
   -- raise a Program_Error) and which was marked as flag option kind
   -- at the time of its declaration.

   -- ========================================================================
   -- 6. Text formating concerns
   -- ------------------------------------------------------------------------

   Dash                     : constant Character := '-';
   Option_Short_Name_Prefix : constant String (1 .. 1) := String'(1 => Dash);
   Option_Long_Name_Prefix  : constant String (1 .. 2) := Dash & Dash;


   function Option_Short_Name_Text (Option_Name : Character) return String;
   --|Ensures: Result'Length >= 1
   -- Result contains Option_Name

   function Option_Long_Name_Text (Option_Name : String) return String;
   --|Requires: Option_Name'Length > 1
   --|Ensures: Result'Length >= Option_Name'Length
   -- Result contains Option_Name as a substring

   -- ========================================================================
   -- ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CR
   -- ========================================================================

private

   type Option_Type;

   type Option_Access is access Option_Type;

   type Option_Accessor_Type  is
   abstract tagged
     null record;

   Null_Option : constant Option_Access := null;

end Shell_UI.Command_Line_Options;
