
package body Shell_UI.Formated_Strings is

   package Texts renames Tiny.Texts;
   package IO renames Ada.Text_IO;

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Formating types
   -- 2. Formating environment : no implementation
   -- 3. General purpose formating
   -- 4. Rudimentary semantic formating
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. Formating types
   -- ------------------------------------------------------------------------

   function Add
     (Position : Position_Type;
      Offset   : Line_Length_Type)
      return Position_Type
   is
   begin
      return Position_Type (Line_Length_Type (Position) + Offset);
   end Add;

   -- ------------------------------------------------------------------------

   function To_String_Length
     (Length : Line_Length_Type)
      return String_Length_Type
   is
   begin
      return String_Length_Type (Length);
   end To_String_Length;

   -- ------------------------------------------------------------------------

   function To_String_Position
     (Position : Position_Type)
      return String_Position_Type
   is
   begin
      return String_Position_Type (Position);
   end To_String_Position;

   -- ------------------------------------------------------------------------

   function Substract
     (High : Position_Type;
      Low  : Position_Type)
      return String_Length_Type
   is
   begin
      return To_String_Position (High) - To_String_Position (Low);
   end Substract;

   -- ========================================================================
   -- 3. General purpose formating
   -- ------------------------------------------------------------------------
   -- 3.1 Section separator
   -- 3.2 Fields padding
   -- 3.3 Laid on N fields
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 3.1 Section separator
   -- ------------------------------------------------------------------------

   Section_Separator_Unit : constant Character := '_';
   Section_Separator_Unit_Is_Low_On_The_Line : constant Boolean := True;

   Section_Separator : constant String
     (1 .. Positive (Console_Line_Width)) :=
     (String'(1 .. Console_Line_Width => Section_Separator_Unit));

   procedure Put_Section_Separator is
   begin
      IO.Put_Line (Section_Separator);
      if Section_Separator_Unit_Is_Low_On_The_Line then
         IO.New_Line;
      end if;
   end Put_Section_Separator;

   -- ========================================================================
   -- 3.2 Fields padding
   -- ------------------------------------------------------------------------

   function Fields_Paddings_For_Fields_Layouts
     (Contents_Lengths : Strings_Lengths_Type;
      Fields_Widths    : Strings_Lengths_Type)
      return Paddings_Type
   is
   begin
      Check_Validity_Constraints :
      begin
         if Contents_Lengths'Length /= Fields_Widths'Length then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;

      Check_Special_Cases :
      begin
         if Contents_Lengths'Length = 0 then
            return Result : constant Paddings_Type (1 .. 0) :=
              (Paddings_Type'(1 .. 0 => 0))
            do
               null;
            end return;
         end if;
      end Check_Special_Cases;

      None_Empty_List :
      declare
         Paddings : Paddings_Type (Fields_Widths'Range);
         -- Function result
         Carry    : String_Length_Type := 0;
         -- See comments in Add_Paddings
      begin
         Give_Paddings_As_Required :
         declare
            J : Positive := Fields_Widths'First;
            -- Index in Fields_Widths and Paddings (boths have same range)
         begin
            -- The basic rule is that a field is given a Padding whose size is
            -- the difference between the Content_Length and the Field_Width.
            -- If a Content_Length overflows the Field_Width, the Padding is
            -- set to zero and the size in excess is set/added to Carry. If a
            -- field is to recieve a Padding, the size in Carry is first
            -- removed from the Padding as possible and the size substracted
            -- to Padding is also substracted to Carry.
            Give_Paddings_As_Required_And_Available :
            for I in Contents_Lengths'Range loop
               declare
                  Content_Length : constant String_Length_Type :=
                    (Contents_Lengths (I));
                  Field_Width    : constant String_Length_Type :=
                    (Fields_Widths (J));
               begin
                  if Field_Width >= Content_Length then
                     declare
                        Padding : String_Length_Type :=
                          (Field_Width - Content_Length);
                     begin
                        if Padding >= Carry then
                           Padding := Padding - Carry;
                           Carry := 0;
                           Paddings (J) := Padding;
                        else
                           Carry := Carry - Padding;
                           Padding := 0;
                           Paddings (J) := Padding;
                        end if;
                     end;
                  else
                     Paddings (J) := 0;
                     Carry := Carry + (Content_Length - Field_Width);
                  end if;
               end;
               if I < Contents_Lengths'Last then
                  J := J + 1;
               end if;
            end loop Give_Paddings_As_Required_And_Available;
         end Give_Paddings_As_Required;

         if Carry = 0 then
            return Paddings;
         end if;

         Reduce_Remaining_Carry_As_Possible :
         begin
            -- Whenever a Content_Length exceeded its allocated Field_Width,
            -- the latter loop was reporting Padding reduction as much as
            -- possible to numbers of successor fields using Carry, up to
            -- get a zero Carry. But the Carry may still be non-zero. If it
            -- is, we try to go backward to report Padding reduction to
            -- predecessor fields as well.
            Reduce_Paddings_Backward :
            for I in reverse Paddings'Range loop
               declare
                  Padding : String_Length_Type := Paddings (I);
               begin
                  if Padding > 0 then
                     if Padding >= Carry then
                        Padding := Padding - Carry;
                        Carry := 0;
                        Paddings (I) := Padding;
                        exit Reduce_Paddings_Backward;
                     else
                        Carry := Carry - Padding;
                        Padding := 0;
                        Paddings (I) := Padding;
                     end if;
                  end if;
               end;
            end loop Reduce_Paddings_Backward;
         end Reduce_Remaining_Carry_As_Possible;

         return Paddings;
         -- Carry may still be non-zero. If it is, then all paddings
         -- will be zero.

      end None_Empty_List;

   end Fields_Paddings_For_Fields_Layouts;

   -- ========================================================================
   -- 3.3 Laid on N fields
   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_Two_Fields
     (Separator : in Character;
      Content_1 : in String;
      Border_1  : in Position_Type;
      Content_2 : in String;
      Border_2  : in Position_Type)
   is
   begin
      Check_Validity_Contraints :
      begin
         if not (Border_2 > Border_1) then
            raise Program_Error;
         end if;
      end Check_Validity_Contraints;

      declare
         Field_1_Width    : constant String_Length_Type :=
           Substract (High => Border_1, Low => IO.Col);
         Field_2_Width    : constant String_Length_Type :=
           Substract (High => Border_2, Low => Border_1 + 1);
         -- Border_2 is guaranted to be greater than Border_1, so Border_1 + 1
         -- is guaranted to not commit any overflow (as long as Border_2 is
         -- valid).
         Contents_Lengths : constant Strings_Lengths_Type (1 .. 2) :=
           (1 => Content_1'Length,
            2 => Content_2'Length);
         Fields_Widths    : constant Strings_Lengths_Type (1 .. 2) :=
           (1 => Field_1_Width,
            2 => Field_2_Width);
         Paddings         : constant Paddings_Type (1 .. 2) :=
           (Fields_Paddings_For_Fields_Layouts
              (Contents_Lengths => Contents_Lengths,
               Fields_Widths    => Fields_Widths));
      begin
         IO.Put (Content_1);
         IO.Put (Texts.Blank_String (Paddings (1)));
         IO.Put (Separator);
         IO.Put (Content_2);
         IO.Put (Texts.Blank_String (Paddings (2)));
      end;

   end Put_Laid_On_Two_Fields;

   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_Three_Fields
     (Separator : in Character;
      Content_1 : in String;
      Border_1  : in Position_Type;
      Content_2 : in String;
      Border_2  : in Position_Type;
      Content_3 : in String;
      Border_3  : in Position_Type)
   is
   begin
      Check_Validity_Contraints :
      begin
         if not (Border_1 > IO.Col) then
            raise Program_Error;
         end if;
         if not (Border_2 > Border_1) then
            raise Program_Error;
         end if;
         if not (Border_3 > Border_2) then
            raise Program_Error;
         end if;
      end Check_Validity_Contraints;

      declare
         Field_1_Width    : constant String_Length_Type :=
           Substract (High => Border_1, Low => IO.Col);
         Field_2_Width    : constant String_Length_Type :=
           Substract (High => Border_2, Low => Border_1 + 1);
         -- Border_2 is guaranted to be greater than Border_1, so Border_1 + 1
         -- is guaranted to not commit any overflow (as long as Border_2 is
         -- valid).
         Field_3_Width    : constant String_Length_Type :=
           Substract (High => Border_3, Low => Border_2 + 1);
         -- Border_3 is guaranted to be greater than Border_2, so Border_2 + 1
         -- is guaranted to not commit any overflow (as long as Border_3 is
         -- valid).
         Contents_Lengths : constant Strings_Lengths_Type (1 .. 3) :=
           (1 => Content_1'Length,
            2 => Content_2'Length,
            3 => Content_3'Length);
         Fields_Widths    : constant Strings_Lengths_Type (1 .. 3) :=
           (1 => Field_1_Width,
            2 => Field_2_Width,
            3 => Field_3_Width);
         Paddings         : constant Paddings_Type (1 .. 3) :=
           (Fields_Paddings_For_Fields_Layouts
              (Contents_Lengths => Contents_Lengths,
               Fields_Widths    => Fields_Widths));
      begin
         IO.Put (Content_1);
         IO.Put (Texts.Blank_String (Paddings (1)));
         IO.Put (Separator);
         IO.Put (Content_2);
         IO.Put (Texts.Blank_String (Paddings (2)));
         IO.Put (Separator);
         IO.Put (Content_3);
         IO.Put (Texts.Blank_String (Paddings (3)));
      end;

   end Put_Laid_On_Three_Fields;

   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_N_Fields
     (Separator : Character;
      Contents  : Strings_Type;
      Borders   : Columns_Positions_Type)
   is
   begin
      -- =====================================================================
      -- Check Validity_Constraints
      -- Handle Special_Cases
      -- Get_Cumulated_Contents_Lengths
      -- Get_Fields_Widths
      -- Get paddings
      -- Write the laid fields using paddings
      -- ---------------------------------------------------------------------

      Validity_Constraints :
      begin
         if not (Borders'Length = Contents'Length) then
            raise Program_Error;
         end if;
         if not (Contents'Length > 0) then
            raise Program_Error;
         end if;
         -- Contents'Length > 0, so the same for Borders'Length, as they
         -- are equal.
         for I in Borders'First + 1 .. Borders'Last loop
            if not (Borders (I) > Borders (I - 1)) then
               raise Program_Error;
            end if;
         end loop;
      end Validity_Constraints;

      Special_Cases :
      begin
         if Contents'Length = 1 then
            IO.Put ( Contents (Contents'First).all);
         end if;
      end Special_Cases;

      -- List is made up of at least two items.
      Common_Case :
      declare
         function Get_Cumulated_Contents_Lengths return Strings_Lengths_Type
         -- Will be use as a parameter of Fields_Paddings_For_Fields_Layouts
         --|Ensures: Result'Range = Contents'Range
         -- To specification to ease use of the result when iteration is
         -- to be done on both Contents and Result.
         is
         begin
            return Result : Strings_Lengths_Type (Contents'Range) do
               for I in Contents'Range loop
                  Result (I) := Contents (I)'Length;
               end loop;
            end return;

         end Get_Cumulated_Contents_Lengths;

         Cumulated_Contents_Lengths : constant Strings_Lengths_Type :=
           Get_Cumulated_Contents_Lengths;

         function Get_Fields_Widths return Strings_Lengths_Type
         -- An array of each length of each element of Contents.
         -- Will be use as a parameter of Fields_Paddings_For_Fields_Layouts
         --|Ensures: Result'Range = Borders'Range
         -- This specification to ease use of the result when iteration is
         -- to be done on both Borders and Result.
         is
         begin
            return Result : Strings_Lengths_Type (Borders'Range) do
               declare
                  Left_Border  : Position_Type := Position_Type'First;
                  Right_Border : Position_Type;
               begin
                  for I in Borders'Range loop
                     Right_Border := Borders (I);
                     Result (I) := Substract
                       (High => Right_Border,
                        Low  => Left_Border);
                     Left_Border := Right_Border;
                  end loop;
               end;
            end return;
         end Get_Fields_Widths;

         Fields_Widths : constant Strings_Lengths_Type := Get_Fields_Widths;

         Paddings : constant Paddings_Type (1 .. Contents'Length) :=
           (Fields_Paddings_For_Fields_Layouts
              (Contents_Lengths => Cumulated_Contents_Lengths,
               Fields_Widths    => Fields_Widths));
         -- This array of paddings is the concret resolution of the requested
         -- layout. The application of these paddings produces the layout.

      begin
         declare
            J : Positive := Paddings'First;
         begin
            for I in Contents'Range loop
               -- For each field, write the content, the padding and
               -- the right border (the latter, only if the field is
               -- not the last).
               IO.Put (Contents (I).all);
               IO.Put (Texts.Blank_String (Paddings (J)));
               if I < Contents'Last then
                  J := J + 1;
                  IO.Put (Separator);
               end if;
            end loop;
         end;
      end Common_Case;

   end Put_Laid_On_N_Fields;

   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_One_Column
     (Text  : in String;
      First : in Position_Type;
      Last  : in Position_Type)
   is
   begin
      -- =====================================================================
      -- Check Validity_Constraints
      -- Handle Special_Cases
      -- Defines the operation to be applied on each token (word/blank)
      -- Iteratively apply the operation on each token of Text
      -- ---------------------------------------------------------------------

      Validity_Constraints :
      begin
         if not (First < Last) then
            raise Program_Error;
         end if;
         if not (Last <= Console_Line_Last) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Special_Cases :
      begin
         if Text'Length = 0 then
            if IO.Col < First then
               IO.Put
                 (Texts.Blank_String
                    (Substract (High => First, Low => IO.Col)));
            end if;
         end if;

         -- All blank Text :
         -- Would be interesting to check if the Text is all blank, but
         -- it is avoided to preserve performance. The Common_Case must
         -- be designed to work well whenever the Text is all blank.

      end Special_Cases;

      Common_Case :
      declare
         -- As the specification requires the layout never ends with a new
         -- line, and as blank after a new line are required to be dropped,
         -- all new line must be deffered to not output a new line which
         -- be followed by a blank, which could occurs if the Text is
         -- either all blank or ends with a blank.

         type Text_Token_Handler_Type is
           new Texts.Token_Accessor_Type
         -- The type which will handle token iteration on Text
         with record
            Pending_New_Line : Boolean := False;
            -- A new line output was requested but it is deffered. It will
            -- be written only prior the output of a none-blank token.
            -- Initially False, because at the initial state, we do not
            -- make any new line request.
         end record;

         procedure Do_Access
           (Accessor : in out Text_Token_Handler_Type;
            Token    : in String;
            Kind     : in Texts.Token_Kinds)
         -- The method which is invoked on Text_Token_Handler_Type on each
         -- token. This is thus the method which has the main responsability
         -- in applying the layout. Except one : one part of the layout,
         -- before the start of any output, is done by process which run the
         -- iteration.
         is
            Position_In_Line : constant Position_Type := IO.Col;
            -- It is unlikely this could raise an exception,
            -- although formally, this can possibly occurs.
            On_Start_Of_Line : constant Boolean := Position_In_Line = First;
         begin
            if not Accessor.Pending_New_Line then
               if On_Start_Of_Line then
                  case Kind is
                     when Texts.A_Blank =>
                        null;
                     when Texts.A_Word =>
                        IO.Put (Token);
                  end case;
               elsif Position_In_Line + (Token'Length - 1) <= Last then
                  -- Token'Length is always greater than 0.
                  IO.Put (Token);
               else
                  case Kind is
                     when Texts.A_Blank =>
                        Accessor.Pending_New_Line := True;
                     when Texts.A_Word =>
                        IO.New_Line;
                        IO.Put
                          (Texts.Blank_String
                             (Substract (High => First, Low => IO.Col)));
                        IO.Put (Token);
                  end case;
               end if;
            else
               case Kind is
                  when Texts.A_Blank =>
                     -- Should never occurs
                     raise Program_Error;
                  when Texts.A_Word =>
                     IO.New_Line;
                     IO.Put
                       (Texts.Blank_String
                          (Substract (High => First, Low => IO.Col)));
                     Accessor.Pending_New_Line := False;
                     IO.Put (Token);
               end case;
            end if;
         end Do_Access;

         Text_Token_Handler : Text_Token_Handler_Type;

      begin
         if IO.Col < First then
            IO.Put
              (Texts.Blank_String
                 (Substract (High => First, Low => IO.Col)));
         end if;

         Texts.For_Each_Word_And_Blank
           (Source   => Text,
            Accessor => Text_Token_Handler);
      end Common_Case;

   end Put_Laid_On_One_Column;

   -- ------------------------------------------------------------------------

   function Arrow
     (Start_Point : Position_Type;
      End_Point   : Position_Type)
      return String
   --|Requires: Start_Point /= End_Point
   -- Start_Point and End_Point cannot be the same, otherwise the arrow
   -- direction could not be defined.
   --|Ensures:
   --|   Start_Point < End_Point ->
   --|   Result'Length = Start_Point - End_Point + 1
   --|Ensures:
   --|   Start_Point < End_Point ->
   --|   Result (Result'Last) = '>'
   --|Ensures:
   --|   Start_Point > End_Point ->
   --|   Result'Length = End_Point - Start_Point + 1
   --|Ensures:
   --|   End_Point < Start_Point ->
   --|   Result (Result'First) = '<'
   is
   begin
      Validity_Constraints :
      begin
         if not (Start_Point /= End_Point) then
            -- Should never occurs, as the only clients of this methods
            -- is in this implementation.
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      declare
         Arrow_Head_To_Right : constant Character := '>';
         Arrow_Head_To_Left  : constant Character := '<';
         Arrow_Body_Unit     : constant Character := '-';

         String_Start        : String_Position_Type renames
           To_String_Position (Start_Point);
         String_End          : String_Position_Type renames
           To_String_Position (End_Point);
      begin
         if Start_Point < End_Point then
            return
              Result : String (String_Start .. String_End) :=
              String'(String_Start .. String_End => Arrow_Body_Unit)
            do
               Result (Result'Last) := Arrow_Head_To_Right;
            end return;
         else
            return
              Result : String (String_Start .. String_End) :=
              String'(String_Start .. String_End => Arrow_Body_Unit)
            do
               Result (Result'First) := Arrow_Head_To_Left;
            end return;
         end if;
      end Common_Case;

   end Arrow;

   -- ------------------------------------------------------------------------

   procedure Put_Arrow_To (End_Point : Position_Type)
   is
   begin
      Validity_Constraints :
      begin
         if not (End_Point > IO.Col) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      IO.Put (Arrow (Start_Point => IO.Col, End_Point => End_Point));
   end;

   -- ------------------------------------------------------------------------

   procedure Put_Arrow_From (Start_Point : Position_Type)
   is
   begin
      Validity_Constraints :
      begin
         if not (Start_Point > IO.Col) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      IO.Put (Arrow (Start_Point => Start_Point, End_Point => IO.Col));
   end;

   -- ========================================================================
   -- 4. Rudimentary semantic formating
   -- ------------------------------------------------------------------------

   function Type_Mark_Text (Type_Name : String) return String
   is
   begin
      return Open_Type_Mark & Type_Name & Close_Type_Mark;
   end Type_Mark_Text;

   -- ------------------------------------------------------------------------

   function Entity_Name_Text (Entity_Name : String) return String
   is
   begin
      return Open_Entity_Name & Entity_Name & Close_Entity_Name;
   end Entity_Name_Text;

   -- ------------------------------------------------------------------------

   function Keyword_Text (Keyword : String) return String
   is
   begin
      return Open_Keyword & Keyword & Close_Keyword;
   end Keyword_Text;

end Shell_UI.Formated_Strings;
