
-- Strings formating in the intent of a shell application output. Formatings
-- are mostly domain specific. This does formating only and do not write
-- anything (string functions).
-- Formating intended to a text output, have to deal with different
-- lengths and bounds types : the ones of standard String and the ones of
-- Ada.Text_IO (while Ada.Text_IO still use standard String as argument types
-- of its output and input methods). This package is valid (can be elaborated)
-- only if some assertions defined in this specification are checked.
-- Methods computing formating whose result is unlikely to be the argument
-- of an another formating process, is directly written and is made a
-- procedure instead of a function. Other formatings are made function
-- returning String.

with Tiny.Texts;
with Ada.Text_IO;

package Shell_UI.Formated_Strings is

   pragma Elaborate_Body;

   -- ========================================================================
   -- Text table
   -- ------------------------------------------------------------------------
   -- See [3 General purpose formating]
   -- See [5 Rudimentary semantic formating]

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Formating types
   -- 2. Formating environment
   -- 3. General purpose formating
   -- 4. Rudimentary semantic formating
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. Formating types
   -- ------------------------------------------------------------------------

   subtype Position_Type is Ada.Text_IO.Positive_Count;
   use type Position_Type;

   Position_Type_First : constant := Position_Type'First;
   Position_Type_Last  : constant := Position_Type'Last;

   type Line_Length_Type is range
     (0) ..
     (Position_Type_Last - Position_Type_First + 1);
   -- The maximum line length is the maximum length one may get between two
   -- Position.

   function Add
     (Position : Position_Type;
      Offset   : Line_Length_Type)
      return Position_Type;
   --|Requires: Offset <= Position_Type'Last - Position

   subtype Strings_Type         is Tiny.Texts.Strings_Type;
   subtype String_Length_Type   is Tiny.Texts.String_Length_Type;
   subtype String_Position_Type is Tiny.Texts.Position_Type;

   -- The following named numbers turns bounds attributes to
   -- universal integer. These named numbers are used in assert pragmas
   -- which are used to ensure safety of some arithmetics between the
   -- involded types. Indeed, String and Text_IO does not use the same
   -- bound types. We requires the length and position types of standard
   -- strings can at least includes the one of Ada.Text_IO (the reverse is
   -- not required).

   String_Length_Type_First   : constant := String_Length_Type'First;
   String_Length_Type_Last    : constant := String_Length_Type'Last;
   String_Position_Type_First : constant := String_Position_Type'First;
   String_Position_Type_Last  : constant := String_Position_Type'Last;
   Line_Length_Type_First     : constant := Line_Length_Type'First;
   Line_Length_Type_Last      : constant := Line_Length_Type'Last;

   -- The following pragma ensures computation of string length based on
   -- line widths and positions is safe. Please, enable compiler warnings
   -- to get a chance to catch any violation at compile time rather than
   -- at runtime elaboration.

   pragma Assert (String_Length_Type_First   <= Line_Length_Type_First);
   pragma Assert (String_Length_Type_Last    >= Line_Length_Type_Last);
   pragma Assert (String_Position_Type_First <= Position_Type_First);
   pragma Assert (String_Position_Type_Last  >= Position_Type_Last);

   -- After these assertions, we can now define the following :

   function To_String_Length
     (Length : Line_Length_Type)
      return String_Length_Type;

   function To_String_Position
     (Position : Position_Type)
      return String_Position_Type;

   function Substract
     (High : Position_Type;
      Low  : Position_Type)
      return String_Length_Type;
   --|Requires: Low <= High

   -- Finished with the latter section.

   type Strings_Lengths_Type is array
     (Positive range <>) of
     String_Length_Type;

   subtype Paddings_Type is Strings_Lengths_Type;
   subtype Indentations_Type is Strings_Lengths_Type;

   type Columns_Positions_Type is array
     (Positive range <>) of
     Position_Type;

   -- ========================================================================
   -- 2. Formating environment
   -- ------------------------------------------------------------------------

   Console_Line_First : constant := Position_Type'First;

   Console_Line_Width : constant := 78;
   -- 80 is the standard, 80 less two, to be nice with shells which use
   -- some kind of ASCII art frame around.

   Console_Line_Last : constant :=
     (Console_Line_First + (Console_Line_Width - 1));

   subtype Console_Line_Position_Type is Position_Type range
     (Console_Line_First) ..
     (Console_Line_Last);

   -- ========================================================================
   -- 3. General purpose formating
   -- ------------------------------------------------------------------------

   Blank                    : constant Character := Tiny.Texts.Blank;
   Empty_String             : constant String := Tiny.Texts.Empty_String;
   Vertical_Line            : constant Character := '|';

   -- ------------------------------------------------------------------------

   procedure Put_Section_Separator;

   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_Two_Fields
     (Separator : in Character;
      Content_1 : in String;
      Border_1  : in Position_Type;
      Content_2 : in String;
      Border_2  : in Position_Type);
   -- None-desctructive formating of fields : if some fields overflow, then
   -- the middle border may be moved left or right to try to fit overall
   -- width, which means, not changing the position of Border_2 (which may
   -- not always be possible).
   --|Requires: Border_1 > Text_IO.Col
   --|Requires: Border_2 > Border_1
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result'Length >= (Border_2 - Positive'First)
   -- The separator character is added only inside of the layout. It is
   -- not added before the first field nor after the last field.
   -- Any of Content_1 or Content_2 may be an empty string.
   -- The first column of the first field is defined to be Positive'First.

   procedure Put_Laid_On_Three_Fields
     (Separator : in Character;
      Content_1 : in String;
      Border_1  : in Position_Type;
      Content_2 : in String;
      Border_2  : in Position_Type;
      Content_3 : in String;
      Border_3  : in Position_Type);
   -- None-desctructive formating of fields : if some fields overflow, then
   -- borders may be moved left or right to try to fit overall which means,
   -- try to no change the position of Border_3 (which not always be
   -- possible).
   --|Requires: Border_1 > Text_IO.Col
   --|Requires: Border_2 > Border_1
   --|Requires: Border_3 > Border_2
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result'Length >= (Border_3 - Positive'First)
   -- The separator character is added only inside of the layout. It is
   -- not added before the first field nor after the last field.
   -- Any of Content_1, Content_2 or Content_3 may be an empty string.
   -- The first column of the first field is defined to be Positive'First.

   procedure Put_Laid_On_N_Fields
     (Separator : Character;
      Contents  : Strings_Type;
      Borders   : Columns_Positions_Type);
   -- None-desctructive formating of fields : if some fields overflow, then
   -- borders may be moved left or right to try to fit overall which means,
   -- try to no change the position of the right most border (which not always
   -- be possible).
   --|Requires: Borders'Length = Contents'Length
   --|Requires: Contents'Length > 0
   -- Otherwise, no result could be reuturned at all, would make not sense.
   -- Should the specification be revised to return an empty string when
   -- Contents'Length is zero ? Will this make sense ?
   --|Requires:
   --|    Borders (Borders'First) > Text_IO.Col
   --|Requires:
   --|    for each
   --|       I in
   --|         (Borders'First + 1 .. Borders'Last),
   --|       Borders (I) > Borders (I - 1)
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result'Length >= (Borders (Borders'Last) - Positive'First)
   -- The separator character is added only inside of the layout. It is
   -- not added before the first field nor after the last field.
   -- Any of Contents may be an empty string. The first column of the first
   -- field is defined to be Positive'First.

   -- ------------------------------------------------------------------------

   procedure Put_Laid_On_One_Column
     (Text  : in String;
      First : in Position_Type;
      Last  : in Position_Type);
   -- None-destructive column formatting. The followings rules specifies the
   -- applied formating principle. The purpose is to layout a text on a
   -- column in the range First .. Last, but some details are required
   -- to specify the behaviour about blanks, too much long tokens and final
   -- state.
   -- Output is done with Ada.Text_IO using the standard output stream.
   -- About the first line : if at the time of the invocation, Text_IO.Col is
   -- less than First, then a padding naturally added to move to First,
   -- if Text_IO.Col is in First .. Last, output is started at
   -- Text_IO.Col, if Text_IO.Col is beyond Last, output is started
   -- on a new line at column First before the first none-blank token is
   -- written. Simply speaking, none-blank token outputs always starts with a
   -- Text_IO.Col in First .. Last.
   -- Then, about tokens which will overflows from First .. Last :
   -- if a token will overflows the width in Text_IO.Col .. Last,
   -- then it is started on a new line at column First, except if
   -- this is a blank token which will then simply been not written, and if
   -- then token still overflow the width in First .. Last, it is still
   -- written there (so the line may wrap if output goes to a console, but
   -- it will not wrap if output goes to a text file via shell I/O
   -- redirection).
   -- About blanks : blank tokens are written as-is, and not compacted, unless
   -- the blank token is at the begining of a new line, in which case, it will
   -- not be written.
   -- About a blank token as the first item of the Text : output start,
   -- whatever is the position of Text_IO.Col, is treated as the start of
   -- a new line, so any initial blank token will be dropped.
   --|Requires: First < Last
   --|Requires: Last <= Screen_Line_Last
   -- The output never ends with an empty new line.
   -- The output may start on a new line without writing on the initial line.
   -- The latter is so that consecutive invokations of this method on
   -- consecutive text, will produce a layout similar the one it would have
   -- been with a single invokation of the method using all strings
   -- concatenated. The previous first line will not be written on if
   -- Text_IO.Col is beyond First and the first token would commit an
   -- overflow, just like it should be inside of the layout this method
   -- applies.
   --|Ensures: Text_IO.Col >= First
   -- In most of cases, Text_IO.Col will be beyong First, but it may be
   -- equal to First, if the Text contains only blanks.
   -- Except with blanks which are ommited at the start of a new line or at
   -- the start of the first line, no truncation is done during formating, and
   -- no token is splitted, thus Text content is preserved and not altered by
   -- formatting.
   -- Think about the range 1 .. First - 1 as a margin and about the range
   -- First .. Last as the column width.

   -- ------------------------------------------------------------------------

   procedure Put_Arrow_To (End_Point : Position_Type);
   --|Requires: End_Point /= Ada.Text_IO.Col
   --|Requires: End_Point > Ada.Text_IO.Col

   procedure Put_Arrow_From (Start_Point : Position_Type);
   --|Requires: Start_Point /= Ada.Text_IO.Col
   --|Requires: Start_Point > Ada.Text_IO.Col

   -- ========================================================================
   -- 4. Rudimentary semantic formating
   -- ------------------------------------------------------------------------

   Key_Value_Separator      : constant String := " : ";
   Argument_Separator       : constant String := ", ";
   Alternative_Operator     : constant String := " | ";
   Copyright_Symbol         : constant String := "(c)";
   Open_Type_Mark           : constant Character := '<';
   Close_Type_Mark          : constant Character := '>';
   Open_Entity_Name         : constant Character := '<';
   Close_Entity_Name        : constant Character := '>';
   Open_Keyword             : constant Character := '<';
   Close_Keyword            : constant Character := '>';
   Major_Minor_Separator    : constant Character := '.';

   -- ------------------------------------------------------------------------

   function Type_Mark_Text (Type_Name : String) return String;
   --|Ensures: Result'Length >= Type_Name'Length
   -- Result contains Type_Name as a substring

   function Entity_Name_Text (Entity_Name : String) return String;
   --|Ensures: Result'Length >= Entity_Name'Length
   -- Result contains Entity_Name as a substring

   function Keyword_Text (Keyword : String) return String;
   --|Ensures: Result'Length >= Keyword'Length
   -- Result contains Keyword as a substring

   -- ========================================================================
   -- ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CR
   -- ========================================================================

private

   pragma Inline (To_String_Length);
   pragma Inline (To_String_Position);

end Shell_UI.Formated_Strings;
