
with Ada.Text_IO;
with Shell_UI.Application_Description;
with Shell_UI.Command_Line_Options;
with Shell_UI.Formated_Strings;

pragma Elaborate_All (Shell_UI.Formated_Strings);

package body Shell_UI.Help is

   package IO          renames Ada.Text_IO;
   package Description renames Shell_UI.Application_Description;
   package Options     renames Shell_UI.Command_Line_Options;
   package Formating   renames Shell_UI.Formated_Strings;

   -- ========================================================================
   -- Text table
   -- ------------------------------------------------------------------------
   -- See [1.1 Indicator text]

   -- ========================================================================
   -- 1. Display help/usage and informations
   -- ------------------------------------------------------------------------
   -- 1.1 Indicator text
   -- 1.2 Options help layout measurements
   -- 1.3 Copyrights
   -- 1.4 Version
   -- 1.5 A-Propos
   -- 1.6 Informations
   -- 1.7 Usage help
   -- 1.8 Usage and informations
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1.1 Indicator text
   -- ------------------------------------------------------------------------

   Flag_Indicator_Text                : constant String := "(flag)";
   Version_Indicator_Text             : constant String := "Version";
   Copyright_Indicator_Text           : constant String := "Copyright";
   Version_Alpha_Indicator_Text       : constant String := "Alpha";
   Version_Beta_Indicator_Text        : constant String := "Beta";
   Version_Pre_Release_Indicator_Text : constant String := "pre-Release";
   Version_Release_Indicator_Text     : constant String := "Release";
   E_Mail_Indicator_Text              : constant String := "Contact";
   Error_Message_Indicator_Text       : constant String := "Error";

   Ada_Application_Point              : constant String :=
     "(an Ada 2005 application)";

   No_Parameters_Indicator_Text       : constant String :=
     "This program expects no parameters";

   -- ========================================================================
   -- 1.2 Options help layout measurements
   -- ------------------------------------------------------------------------

   Short_Name_Right_Border      : constant := 3;
   Long_Name_Right_Border       : constant := 16;
   Type_Indication_Right_Border : constant := 25;

   Left_Margin_Length : constant Formating.Line_Length_Type := 27;

   Right_After_Left_Margin : constant Formating.Position_Type :=
     Formating.Add
       (Position => Formating.Console_Line_First,
        Offset   => Left_Margin_Length);

   Left_Margin_Last : constant Formating.Position_Type :=
     Formating.Position_Type'Pred (Right_After_Left_Margin);

   Console_Line_Last : constant Formating.Position_Type :=
     Formating.Console_Line_Last;

   -- ========================================================================
   -- 1.3 Copyrights
   -- ------------------------------------------------------------------------

   procedure Display_Copyrights
   is
   begin
      Validity_Constraints :
      begin
         if not Description.Name_Fixed then
            raise Program_Error;
         end if;
         if not Description.Main_Copyright_Notice_Fixed then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      IO.New_Line;

      IO.Put (Copyright_Indicator_Text);
      IO.Put (Formating.Key_Value_Separator);
      IO.Put (Description.Name);
      IO.Put (Formating.Argument_Separator);
      IO.Put (Formating.Copyright_Symbol);
      IO.Put (Formating.Blank);
      IO.Put (Description.Main_Copyright_Notice);
      IO.New_Line;

      if Description.There_Are_Third_Parties_Copyrights then
         declare
            type Third_Party_Copyright_Accessor_Type
              is new Description.Third_Party_Copyright_Accessor_Type with
                null record;

            procedure Do_Access
              (Accessor         : in out Third_Party_Copyright_Accessor_Type;
               Component_Name   : in String;
               Copyright_Notice : in String)
            is
            begin
               pragma Unreferenced (Accessor);
               IO.Put (Copyright_Indicator_Text);
               IO.Put (Formating.Key_Value_Separator);
               IO.Put (Component_Name);
               IO.Put (Formating.Argument_Separator);
               IO.Put (Formating.Copyright_Symbol);
               IO.Put (Formating.Blank);
               IO.Put (Copyright_Notice);
               IO.New_Line;
            end Do_Access;

            Third_Party_Copyright_Accessor
              : Third_Party_Copyright_Accessor_Type;
         begin
            Description.For_Each_Third_Party_Copyright
              (Accessor => Third_Party_Copyright_Accessor);
         end;
      end if;

   end Display_Copyrights;

   -- ========================================================================
   -- 1.4 Version
   -- ------------------------------------------------------------------------

   procedure Display_Application_Version
   is
      function Minus_Left_Sign (S : String) return String
      is
      begin
         if S'Length = 0 then
            return S;
         end if;

         return S (Positive'Succ (S'First) .. S'Last);
      end;
   begin
      Check_Validity_Constraints :
      begin
         if not Description.Version_Fixed then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;

      IO.Put (Description.Name);
      IO.Put (Formating.Blank);
      IO.Put (Version_Indicator_Text);
      IO.Put (Formating.Blank);
      IO.Put (Minus_Left_Sign (Natural'Image (Description.Version_Major)));
      IO.Put (Formating.Major_Minor_Separator);
      IO.Put (Minus_Left_Sign (Natural'Image (Description.Version_Minor)));
      IO.Put (Formating.Blank);

      case Description.Version_Status is
         when Description.Alpha =>
            IO.Put (Version_Alpha_Indicator_Text);
         when Description.Beta =>
            IO.Put (Version_Beta_Indicator_Text);
         when Description.Pre_Release =>
            IO.Put (Version_Pre_Release_Indicator_Text);
         when Description.Release =>
            IO.Put (Version_Release_Indicator_Text);
      end case;

      IO.Put (Formating.Blank);
      IO.Put (Ada_Application_Point);

      IO.New_Line;

   end Display_Application_Version;

   -- ========================================================================
   -- 1.5 A-Propos
   -- ------------------------------------------------------------------------

   procedure Display_A_Propos
   is
   begin
      Check_Validity_Constraints :
      begin
         if not Description.Name_Fixed then
            raise Program_Error;
         end if;

         if not Options.Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;

      Put_Header :
      begin
         IO.New_Line;
         Formating.Put_Section_Separator;
      end Put_Header;

      Put_Purpose :
      begin
         if Description.Purpose_Text_Fixed then
            IO.Put (Description.Name);
            IO.Put (Formating.Key_Value_Separator);
            IO.Put (Description.Purpose_Text);
            IO.New_Line;
            IO.New_Line;
         end if;
      end Put_Purpose;

      Put_Contact :
      begin
         IO.Put (E_Mail_Indicator_Text);
         IO.Put (Formating.Key_Value_Separator);
         IO.Put (Description.Contact_E_Mail);
         IO.New_Line;
         IO.New_Line;
      end Put_Contact;

      Put_Version_Informations :
      begin
         if Description.Version_Fixed then
            Display_Application_Version;
         end if;
      end Put_Version_Informations;

      Put_Copyrights :
      begin
         if Description.Main_Copyright_Notice_Fixed then
            Display_Copyrights;
         end if;
      end Put_Copyrights;

      Put_Footer :
      begin
         Formating.Put_Section_Separator;
      end Put_Footer;

   end Display_A_Propos;

   -- ========================================================================
   -- 1.6 Application informations
   -- ------------------------------------------------------------------------

   procedure Display_Informations
   is
   begin
      Display_A_Propos;
   end Display_Informations;

   -- ========================================================================
   -- 1.7 Usage help
   -- ------------------------------------------------------------------------
   -- 1.7.1 Before parameters help is displayed
   -- 1.7.2 When there are parameters
   -- 1.7.3 When there are no parameters
   -- 1.7.4 After parameters help was displayed
   -- ------------------------------------------------------------------------

   procedure Display_Usage (Message : in String := Tiny.Texts.Empty_String)
   is
   begin
      Check_Validity_Constraints :
      begin
         if not Options.Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;

      Put_Parameters_Descriptions :
      begin

         Put_Header :
         begin
            -- ===============================================================
            -- 1.7.1 Before parameters help is displayed
            -- ---------------------------------------------------------------

            Formating.Put_Section_Separator;

         end Put_Header;

         if not Options.Options_List_Is_Empty then

            -- ===============================================================
            -- 1.7.2 When there are parameters
            -- ---------------------------------------------------------------
            -- 1.7.2.1 Option accessor which displays each option
            -- 1.7.2.2 Short name field content
            -- 1.7.2.3 Long name field content
            -- 1.7.2.4 Type indication field content
            -- 1.7.2.5 Display the three fields formated
            -- 1.7.2.6 Display the option help text formated
            -- ---------------------------------------------------------------

            Options_List_Not_Empty :
            declare

               -- -------------------------------------
               -- 1.7.2.1 Option accessor which displays each option
               -- -------------------------------------

               type Option_Display_Method_Type is
                 new Options.Option_Accessor_Type
               with null record;

               -- -------------------------------------

               procedure Do_Access
                 (Accessor : in out Option_Display_Method_Type;
                  Element  : in Options.Option_Access)
               is
                  Option : Options.Option_Access renames Element;

                  -- -------------------------------------
                  -- 1.7.2.2 Short name field content
                  -- -------------------------------------

                  function Short_Name_Field_Content return String
                  is
                  begin
                     if
                       Options.Short_Name (Option) /=
                       Options.Option_Null_Short_Name
                     then
                        return
                          Options.Option_Short_Name_Text
                            (Options.Short_Name (Option));
                     else
                        return Formating.Empty_String;
                     end if;
                  end Short_Name_Field_Content;

                  -- -------------------------------------
                  -- 1.7.2.3 Long name field content
                  -- -------------------------------------

                  function Long_Name_Field_Content return String
                  is
                  begin
                     if
                       Options.Long_Name (Option) /=
                       Options.Option_Null_Long_Name
                     then
                        return
                          Options.Option_Long_Name_Text
                            (Options.Long_Name (Option));
                     else
                        return Formating.Empty_String;
                     end if;
                  end Long_Name_Field_Content;

                  -- -------------------------------------
                  -- 1.7.2.4 Type indication field content
                  -- -------------------------------------

                  function Type_Indication_Field_Content return String
                  is
                  begin
                     if not Options.Is_A_Flag (Option) then
                        return
                          Formating.Type_Mark_Text
                            (Options.Type_Name (Option));
                     else
                        return Flag_Indicator_Text;
                     end if;
                  end Type_Indication_Field_Content;

                  -- -------------------------------------

               begin
                  pragma Unreferenced (Accessor);

                  Put_Option_Definition :
                  begin
                     -- -------------------------------------
                     -- 1.7.2.5 Display the three fields formated
                     -- -------------------------------------
                     Formating.Put_Laid_On_Three_Fields
                       (Separator => Formating.Blank,
                        Content_1 => Short_Name_Field_Content,
                        Border_1  => Short_Name_Right_Border,
                        Content_2 => Long_Name_Field_Content,
                        Border_2  => Long_Name_Right_Border,
                        Content_3 => Type_Indication_Field_Content,
                        Border_3  => Type_Indication_Right_Border);
                  end Put_Option_Definition;

                  Put_Option_Help_Text :
                  begin
                     -- -------------------------------------
                     -- 1.7.2.6 Display the option help text formated
                     -- -------------------------------------

                     IO.Put (Formating.Key_Value_Separator);

                     Formating.Put_Laid_On_One_Column
                       (Text  => Options.Help_Text (Option),
                        First => Right_After_Left_Margin,
                        Last  => Console_Line_Last);

                     IO.New_Line;

                  end Put_Option_Help_Text;

                  IO.New_Line;

               end Do_Access;

               -- -------------------------------------

               Option_Display_Method : Option_Display_Method_Type;

            begin

               Options.Read_Each_Options
                 (Accessor => Option_Display_Method);

            end Options_List_Not_Empty;
         else
            -- ===============================================================
            -- 1.7.3 When there are no parameters
            -- ---------------------------------------------------------------

            IO.Put (No_Parameters_Indicator_Text);
            IO.New_Line;
         end if;

         -- ==================================================================
         -- 1.7.4 After parameters help was displayed
         -- ------------------------------------------------------------------

         Contextual_Message :
         begin
            if Message'Length > 0 then

               IO.New_Line;

               IO.Put (Formating.Blank);

               Formating.Put_Arrow_To
                 (Formating.Position_Type'Pred (Left_Margin_Last));

               Formating.Put_Laid_On_One_Column
                 (Text  =>
                    (Error_Message_Indicator_Text &
                     Formating.Key_Value_Separator),
                  First => Right_After_Left_Margin,
                  Last  => Console_Line_Last);

               Formating.Put_Laid_On_One_Column
                 (Text  => Message,
                  First => Right_After_Left_Margin,
                  Last  => Console_Line_Last);

               IO.New_Line;
               IO.New_Line;

            end if;
         end Contextual_Message;

         Put_Footer :
         begin
            Formating.Put_Section_Separator;
         end Put_Footer;

      end Put_Parameters_Descriptions;

   end Display_Usage;

   -- ========================================================================
   -- 1.8 Usage and informations
   -- ------------------------------------------------------------------------

   procedure Display_Usage_And_Informations
   is
   begin
      Check_Validity_Constraints :
      begin
         if not Description.Name_Fixed then
            raise Program_Error;
         end if;

         if not Options.Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Check_Validity_Constraints;


      Put_Header :
      begin
         IO.New_Line;
         Formating.Put_Section_Separator;
         IO.Put (Description.Name);

         if Description.Purpose_Text_Fixed then
            IO.Put (Formating.Key_Value_Separator);
            IO.Put (Description.Purpose_Text);
         end if;

         IO.New_Line;
      end Put_Header;


      Put_Usage_Information :
      begin
         Display_Usage;
      end Put_Usage_Information;

      Put_Footer :
      begin
         if Description.Version_Fixed then

            IO.Put (E_Mail_Indicator_Text);
            IO.Put (Formating.Key_Value_Separator);
            IO.Put (Description.Contact_E_Mail);
            IO.New_Line;
            IO.New_Line;

            Display_Application_Version;

            if Description.Main_Copyright_Notice_Fixed then
               Display_Copyrights;
            end if;

            Formating.Put_Section_Separator;
         end if;
      end Put_Footer;

   end Display_Usage_And_Informations;

end Shell_UI.Help;
