
with Shell_UI.Command_Line_Options;
with Shell_UI.Help;
with Ada.Text_IO;

package body Shell_UI.Standard_Command_Line_Options is

   package CLO  renames Shell_UI.Command_Line_Options;
   package Help renames Shell_UI.Help;
   package IO   renames Ada.Text_IO;

   -- ========================================================================
   -- Text table
   -- ------------------------------------------------------------------------

   Flag_Type_Name : constant String := "flag";

   Help_Option_Short_Name : constant Character := 'h';
   Help_Option_Long_Name : constant String := "help";
   Help_Option_Help : constant String :=
     ("Display command line options help, application purpose and " &
      "version information and then exit.");

   Version_Option_Short_Name : constant Character := 'v';
   Version_Option_Long_Name : constant String := "version";
   Version_Option_Help : constant String :=
     ("Display version informations and go-on with normal processing.");

   Infos_Option_Long_Name : constant String := "infos";
   Infos_Option_Help : constant String :=
     ("Display application purpose and version informations and " &
      "then exit. This option does not provides command line " &
      "options help.");

   A_Propos_Option_Long_Name : constant String := "apropos";
   A_Propos_Option_Help : constant String :=
     ("Display a-propos informations, which includes copyrights and " &
      "then go-on with normal processing.");

   Copyrights_Option_Long_Name : constant String := "copyrights";
   Copyrights_Option_Help : constant String :=
     ("Display copyrights informations and go-on with normal " &
      "processing.");

   Help_Option       : CLO.Option_Access := CLO.Null_Option;
   Version_Option    : CLO.Option_Access := CLO.Null_Option;
   Infos_Option      : CLO.Option_Access := CLO.Null_Option;
   A_Propos_Option   : CLO.Option_Access := CLO.Null_Option;
   Copyrights_Option : CLO.Option_Access := CLO.Null_Option;

   -- So far, all standard options are flag options. Whenever one is not
   -- to be a flag option, then the two function Standard_Option, which
   -- are responsible of registering standard options, will have ot be updated
   -- to include this case.

   function Available
     (Short_Name : Character)
         return Boolean
   is
   begin
      return not CLO.Option_Short_Name_Already_Used (Short_Name);
   end Available;

   function Available
     (Long_Name : String)
         return Boolean
   is
   begin
      return not CLO.Option_Long_Name_Already_Used (Long_Name);
   end Available;

   function Both_Available
     (Short_Name : Character;
      Long_Name  : String)
         return Boolean
   is
   begin
      return Available (Short_Name) and Available (Long_Name);
   end Both_Available;

   function Standard_Option
     (Short_Name : Character;
      Long_Name  : String;
      Help_Text  : String)
      return CLO.Option_Access
   is
   begin
      if Both_Available
        (Short_Name => Short_Name,
         Long_Name  => Long_Name)
      then
         return CLO.Declare_Option
           (Short_Name => Short_Name,
            Long_Name  => Long_Name,
            Is_A_Flag  => True,
            Type_Name  => Flag_Type_Name,
            Help_Text  => Help_Text);
      elsif Available (Short_Name) then
         return CLO.Declare_Option
           (Short_Name => Short_Name,
            Long_Name  => CLO.Option_Null_Long_Name,
            Is_A_Flag  => True,
            Type_Name  => Flag_Type_Name,
            Help_Text  => Help_Text);
      elsif Available (Long_Name) then
         return CLO.Declare_Option
           (Short_Name => CLO.Option_Null_Short_Name,
            Long_Name  => Long_Name,
            Is_A_Flag  => True,
            Type_Name  => Flag_Type_Name,
            Help_Text  => Help_Text);
      else
         return CLO.Null_Option;
      end if;
   end Standard_Option;

   function Standard_Option
     (Long_Name : String;
      Help_Text : String)
      return CLO.Option_Access
   is
   begin
      if Available (Long_Name) then
         return CLO.Declare_Option
           (Short_Name => CLO.Option_Null_Short_Name,
            Long_Name  => Long_Name,
            Is_A_Flag  => True,
            Type_Name  => Flag_Type_Name,
            Help_Text  => Help_Text);
      else
         return CLO.Null_Option;
      end if;
   end Standard_Option;

   procedure Add_Standard_Options
   is
   begin
      Validity_Constraints :
      begin
         if CLO.Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Help_Option := Standard_Option
        (Short_Name => Help_Option_Short_Name,
         Long_Name  => Help_Option_Long_Name,
         Help_Text  => Help_Option_Help);

      Version_Option := Standard_Option
        (Short_Name => Version_Option_Short_Name,
         Long_Name  => Version_Option_Long_Name,
         Help_Text  => Version_Option_Help);

      Infos_Option := Standard_Option
        (Long_Name  => Infos_Option_Long_Name,
         Help_Text  => Infos_Option_Help);

      A_Propos_Option := Standard_Option
        (Long_Name  => A_Propos_Option_Long_Name,
         Help_Text  => A_Propos_Option_Help);

      Copyrights_Option := Standard_Option
        (Long_Name  => Copyrights_Option_Long_Name,
         Help_Text  => Copyrights_Option_Help);

   end;

   procedure First_Handle_Standard_Options
   is
   begin
      Validity_Constraints :
      begin
         if not CLO.Command_Line_Parsed then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      declare
         Show_Help        : constant Boolean := CLO.Flagged (Help_Option);
         Show_Version     : Boolean := CLO.Flagged (Version_Option);
         Show_Infos       : Boolean := CLO.Flagged (Infos_Option);
         Show_A_Propos    : Boolean := CLO.Flagged (A_Propos_Option);
         Show_Copyrights  : Boolean := CLO.Flagged (Copyrights_Option);
         Then_Quit        : Boolean := False;
      begin

         -- Order of tests is important : ordered from included to including.

         if Show_Infos and Show_Version then
            Show_Version := False;
            -- Showing informations includes display of version.
         end if;

         if Show_Infos and Show_A_Propos then
            Show_A_Propos := False;
            -- Showing informations includes display of a-propos
         end if;

         if Show_Infos and Show_Copyrights then
            Show_Copyrights := False;
            -- Showing informations includse display of copyrigths
         end if;

         if Show_Help and Show_Infos then
            Show_Infos := False;
            -- Showing help includse display of infos
         end if;

         -- ------------------------------------------------------------------

         -- Order of tests is not significant

         if Show_Help then
            Then_Quit := True;
         end if;

         if Show_Infos then
            Then_Quit := True;
         end if;

         -- ------------------------------------------------------------------

         if Show_Help then
            Help.Display_Usage_And_Informations;
         end if;

         if Show_Version then
            IO.New_Line;
            Help.Display_Application_Version;
         end if;

         if Show_Infos then
            Help.Display_Informations;
         end if;

         if Show_A_Propos then
            Help.Display_A_Propos;
         end if;

         if Show_Copyrights then
            Help.Display_Copyrights;
         end if;

         -- ------------------------------------------------------------------

         if Then_Quit then
            raise Quit_Application;
         end if;

      end;
   end;

end Shell_UI.Standard_Command_Line_Options;
