
-- On purpose rudimentary vector type : no removal and no anticipation of
-- storage management, only appending. To be used when Ada.Containers.Vectors
-- is too much. Use an active iterators driving tagged types, instead of a
-- Cursor.

package Tiny.Containers is

   -- ------------------------------------------------------------------------

   generic

      type Element_Type is private;

      -- ---------------------------------------------------------------------

   package Vectors is

      -- =====================================================================
      -- Table of Content
      -- ---------------------------------------------------------------------
      -- 1. Vector type
      -- 2. Tester object
      -- 3. Testing iteration
      -- 4. Requester object
      -- 5. Request iteration
      -- 6. Updater object
      -- 7. Update iteration
      -- ---------------------------------------------------------------------

      -- =====================================================================
      -- 1. Vector type
      -- ---------------------------------------------------------------------

      type Vector_Access (<>) is private;
      -- Copying a Vector_Access which is /= Null_Vector, shares an
      -- access to a single vector.

      Null_Vector : constant Vector_Access;

      function Is_Null
        (Vector : Vector_Access)
         return Boolean;
      --|Ensures: (Vector = Null_Vector) = Result
      -- To by-pass "=" and "/=" operators missing imports.

      procedure Append_To
        (Vector  : in out Vector_Access;
         Element : in Element_Type);
      --|Ensures: Vector /= Null_Vector

      -- =====================================================================
      -- 2. Tester object
      -- ---------------------------------------------------------------------

      type Test_Type is abstract tagged private;
      -- For iterative tests on a Vector's elements

      subtype Test_Class is Test_Type'Class;

      function Checked
        (Test    : Test_Type;
         Element : Element_Type)
         return Boolean
         is abstract;

      -- =====================================================================
      -- 3. Testing iteration
      -- ---------------------------------------------------------------------

      function Exist_In
        (Vector : Vector_Access;
         Test   : Test_Class)
         return Boolean;
      --|Ensures: (Vector = Null_Vector) -> (Result = False)

      procedure Find_In
        (Vector  : in Vector_Access;
         Test    : in Test_Class;
         Found   : out Boolean;
         Element : out Element_Type);
      --|Ensures: Found is defined
      --|Ensures: (Vector = Null_Vector) -> (Found = False)
      --|Ensures: (Found = False) -> (Element = old Element)
      -- This is better to have it as a procedure for when the element
      -- is not found and nothing is to be returned, then, the output
      -- parameter if left unchanged.

      -- =====================================================================
      -- 4. Requester object
      -- ---------------------------------------------------------------------

      type Accessor_Type is abstract tagged private;
      -- For iterative requests on a Vector's elements

      subtype Accessor_Class is Accessor_Type'Class;

      procedure Do_Access
        (Accessor    : in out Accessor_Type;
         Element     : in Element_Type)
         is abstract;

      -- =====================================================================
      -- 5. Request iteration
      -- ---------------------------------------------------------------------

      procedure For_Each_In
        (Vector   : in Vector_Access;
         Accessor : in out Accessor_Class);

      -- =====================================================================
      -- 6. Updater object
      -- ---------------------------------------------------------------------

      type Modifier_Type is abstract tagged private;
      -- For iterative updates on a vector's elements

      subtype Modifier_Class is Modifier_Type'Class;

      procedure Modify
        (Modifier    : in out Modifier_Type;
         Element     : in out Element_Type)
         is abstract;

      -- =====================================================================
      -- 7. Update iteration
      -- ---------------------------------------------------------------------

      procedure For_Each_In
        (Vector   : in Vector_Access;
         Modifier : in out Modifier_Class);

      -- =====================================================================
      -- ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT
      -- =====================================================================

   private

      type Test_Type     is abstract tagged null record;
      type Accessor_Type is abstract tagged null record;
      type Modifier_Type is abstract tagged null record;

      type Vector_Type;

      type Vector_Access is access Vector_Type;

      Null_Vector : constant Vector_Access := null;

   end Vectors;

end Tiny.Containers;
