
with Ada.Characters.Handling;

package body Tiny.Texts is

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Definitions : no implementation
   -- 2. Tests
   -- 3. Convertions
   -- 4. Edition
   -- 5. Functions
   -- 6. Tokenizer
   -- 7. Variable substitutions
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 2. Tests
   -- ------------------------------------------------------------------------

   function Is_Graphic
     (C : Character)
      return Boolean
   is
   begin
      if C = Blank then
         return False;
      else
         return Ada.Characters.Handling.Is_Graphic (C);
      end if;
   end Is_Graphic;

   -- ------------------------------------------------------------------------

   function Is_Blank
     (C : Character)
      return Boolean
   is
   begin
      return not Is_Graphic (C);
   end Is_Blank;

   -- ------------------------------------------------------------------------

   function Is_Graphic (S : String) return Boolean
   is
   begin
      for I in S'Range loop
         if not Is_Graphic (S (I)) then
            return False;
         end if;
      end loop;
      return True;
   end Is_Graphic;

   -- ------------------------------------------------------------------------

   function Is_Blank (S : String) return Boolean
   is
   begin
      for I in S'Range loop
         if Is_Graphic (S (I)) then
            return True;
         end if;
      end loop;
      return False;
   end Is_Blank;

   -- ========================================================================
   -- 3. Convertions
   -- ------------------------------------------------------------------------

   function To_String (C : Character) return String
   is
   begin
      return String'(1 => C);
   end To_String;

   -- ------------------------------------------------------------------------

   function To_String
     (Source                : Wide_String;
      Replacement_Character : Character := ASCII_Replacement_Character)
      return String
   is
      function To_Character (Wide : Wide_Character) return Character
      is
         Character_Code : constant Integer := Wide_Character'Pos (Wide);
      begin
         if Character_Code > Character'Pos (Character'Last) then
            return Replacement_Character;
         else
            return Character'Val (Character_Code);
         end if;
      end To_Character;
   begin
      return
        Result : String (Source'Range)
      do
         for I in Source'Range loop
            Result (I) := To_Character (Source (I));
         end loop;
      end return;
   end To_String;

   -- ------------------------------------------------------------------------

   function To_Wide_String (Source : String) return Wide_String
   is
      function To_Wide_Character (C : Character) return Wide_Character
      is
      begin
         return Wide_Character'Val (Character'Pos (C));
      end To_Wide_Character;
   begin
      return
        Result : Wide_String (Source'Range)
      do
         for I in Source'Range loop
            Result (I) := To_Wide_Character (Source (I));
         end loop;
      end return;
   end To_Wide_String;

   -- ========================================================================
   -- 4. Edition
   -- ------------------------------------------------------------------------

   procedure Write
     (Text   : in out String;
      Cursor : in out Positive;
      Item   : in String)
   is
   begin
      Validity_Constraints :
      begin
         if not (Text'Length > 0) then
            raise Program_Error;
         end if;
         if not (Cursor in Text'Range) then
            raise Program_Error;
         end if;
         if Item'Length > 0 then
            if not ((Item'Length - 1) <= (Text'Last - Cursor)) then
               raise Program_Error;
            end if;
         end if;
      end Validity_Constraints;

      Special_Cases :
      begin
         if Item'Length = 0 then
            return;
         end if;
      end Special_Cases;

      Common_Case :
      declare
         Slice_Length : constant String_Length_Type := Item'Length;
         Slice_First  : constant Positive := Cursor;
         Slice_Last   : constant Positive := Slice_First + (Slice_Length - 1);
      begin
         Text (Slice_First .. Slice_Last) := Item;
         if Slice_Last < Text'Last then
            Cursor := Cursor + Slice_Length;
         else
            Cursor := Text'First;
         end if;
      end Common_Case;

   end Write;

   -- ------------------------------------------------------------------------

   procedure Write
     (Text   : in out String;
      Cursor : in out Positive;
      Item   : in Character)
   is
   begin
      Validity_Constraints :
      begin
         if not (Text'Length > 0) then
            raise Program_Error;
         end if;
         if not (Cursor in Text'Range) then
            raise Program_Error;
         end if;
      end Validity_Constraints;

      Common_Case :
      begin
         Text (Cursor) := Item;

         if Cursor < Text'Last then
            Cursor := Cursor + 1;
         else
            Cursor := Text'First;
         end if;
      end Common_Case;

   end Write;

   -- ========================================================================
   -- 5. Functions
   -- ------------------------------------------------------------------------

   function Blank_String (Length : Natural) return String
   is
   begin
      return String'(1 .. Length => Blank);
   end Blank_String;

   -- ------------------------------------------------------------------------

   function Each_Appended (Items : Strings_Type) return String
   is
   begin
      Validity_Constraints :
      begin
         null;
         -- Validity constraint check is done while computing
         -- Get_Cumulated_Length later in Common_Case. This is done
         -- there for factorisation and thus, better performances.
      end Validity_Constraints;

      Special_Cases :
      begin
         if Items'Length = 0 then
            return Empty_String;
         end if;
      end Special_Cases;

      Common_Case :
      declare
         function Get_Cumulated_Length return String_Length_Type
         is
         begin
            return
              Result : String_Length_Type := 0
            do
               for I in Items'Range loop
                  Add_Item_Length_And_Check_For_Validity :
                  declare
                     Item_Length : constant String_Length_Type :=
                       Items (I).all'Length;
                  begin
                     if (String_Length_Type'Last - Result) < Item_Length then
                        raise Program_Error;
                     end if;
                     Result := Result + Item_Length;
                  end Add_Item_Length_And_Check_For_Validity;
               end loop;
            end return;
         end Get_Cumulated_Length;

         Cumulated_Length : constant String_Length_Type :=
           Get_Cumulated_Length;

      begin
         return
           Result : String (1 .. Cumulated_Length)
         do
            Write_Each_Item_To_Result :
            declare
               Cursor : Positive := Result'First;
            begin
               for I in Items'Range loop
                  Write
                    (Text   => Result,
                     Cursor => Cursor,
                     Item   => Items (I).all);
               end loop;
            end Write_Each_Item_To_Result;
         end return;
      end Common_Case;

   end Each_Appended;

   -- ========================================================================
   -- 6. Tokenizer
   -- ------------------------------------------------------------------------

   procedure For_Each_Word_And_Blank
     (Source   : in String;
      Accessor : in out Token_Accessor_Class)
   is
      function Token_Kind_From_Character (C : Character) return Token_Kinds
      is
      begin
         if Tiny.Texts.Is_Blank (C) then
            return A_Blank;
         else
            return A_Word;
         end if;
      end Token_Kind_From_Character;
   begin

      Special_Cases :
      begin
         if Source'Length = 0 then
            -- If Source'Length is zero, then getting Start_Of_Token and
            -- End_Of_Token would be invalid and we also would not be able
            -- to get Token_Kind_From_Character.
            return;
         end if;

         if Source'Length = 1 then
            -- If Source'Length is 1, then we may potientally get a numeric
            -- out of range error when computing Source'First + 1. When
            -- the source is a single character, the sole token is the string
            -- itself.
            Do_Access
              (Accessor => Accessor,
               Token    => Source,
               Kind     => Token_Kind_From_Character (Source (Source'First)));
            return;
         end if;
      end Special_Cases;

      Common_Cases :
      declare
         Loop_Start     : Positive;
         Start_Of_Token : Positive;
         Token_Kind     : Token_Kinds;
      begin

         Start_Of_Token := Source'First;
         -- The first character is interpreted before we enter the loop and
         -- the iteration will start at the position next to this one. Do so
         -- because the loop makes use of a previous state. For the first
         -- loop iteration to ghave one (previous state), we first compute
         -- this state based on the first character and start the iteration
         -- at the second character.

         Token_Kind :=
           Token_Kind_From_Character (Source (Start_Of_Token));
         -- Token_Kind defines the previous state.

         Loop_Start := Positive'Succ (Start_Of_Token);

         for I in Loop_Start .. Source'Last loop
            -- We know we've reached the end of a token, when we switch from
            -- one token kind to another.
            Test_Token_Kind_At_I :
            declare
               Next_Token_Kind : constant Token_Kinds :=
                 Token_Kind_From_Character (Source (I));
            begin
               if Next_Token_Kind /= Token_Kind then
                  -- We've switched from a kind of area to another, so
                  -- we've finished with the previous area and start a with a
                  -- new one. Return the previous as a token and initialize
                  -- the new token.
                  Do_Access
                    (Accessor => Accessor,
                     Token    => Source (Start_Of_Token .. Positive'Pred (I)),
                     Kind     => Token_Kind);
                  Start_Of_Token := I;
                  -- Start of the next token
                  Token_Kind := Next_Token_Kind;
                  -- Kind of the next token which will be handled next.
               end if;
            end Test_Token_Kind_At_I;
         end loop;

         Do_Access
           (Accessor => Accessor,
            Token    => Source (Start_Of_Token .. Source'Last),
            Kind     => Token_Kind);
         -- For the last token, we will never get a character beyond which
         -- will trigger the application of this token, as the last
         -- character of the source is also the last character of the token,
         -- and no token follows. The last one is thus to be inconditionally
         -- applied here, after the loop.

      end Common_Cases;

   end For_Each_Word_And_Blank;

   -- ========================================================================
   -- 7. Variable substitutions
   -- ------------------------------------------------------------------------

   function Variable_Message
     (Pattern : String;
      Value   : String)
      return String
   is
      Tag_Name : constant String := "$1"; --! To be moved
   begin
      Special_Cases :
      begin
         if Pattern'Length < Tag_Name'Length then
            return Pattern;
            -- If the Pattern length is less than the "$1" tag, then
            -- it cannot contains the "$1" tag, and thus no substitution
            -- may occurs for the Pattern.
         end if;
      end Special_Cases;

      Common_Case :
      declare

         function Match_At (Position : Positive) return Boolean
         is
         begin
            for I in Tag_Name'Range loop
               declare
                  J : Positive;
               begin
                  J := Position + (I - Tag_Name'First);
                  if Tag_Name (I) /= Pattern (J) then
                     return False;
                  end if;
               end;
            end loop;
            return True;
         end Match_At;

         Loop_Start : constant Positive := Pattern'First;
         Loop_End : constant Positive :=
           (Pattern'Last) - (Tag_Name'Length - 1);
         Result_Length : Natural := Pattern'Length;
      begin
         Compute_Result_Length :
         begin
            for I in Loop_Start .. Loop_End loop
               if Match_At (I) then
                  Result_Length :=
                    (Result_Length) -
                    (Tag_Name'Length) +
                    (Value'Length);
               end if;
            end loop;
         end Compute_Result_Length;

         Return_Result :
         declare
            Slice_Start      : Positive := Pattern'First;
            Cursor           : Positive := 1;
            Valid_Last_Slice : Boolean := True;
         begin
            return
              Result : String (1 .. Result_Length)
            do
               for I in Loop_Start .. Loop_End loop
                  if Match_At (I) then
                     Texts.Write
                       (Text   => Result,
                        Cursor => Cursor,
                        Item   => Pattern (Slice_Start .. I - 1));
                     Texts.Write
                       (Text   => Result,
                        Cursor => Cursor,
                        Item   => Value);
                     if I < Pattern'Last - (Tag_Name'Length) then
                        Slice_Start := I + Tag_Name'Length;
                     else
                        Valid_Last_Slice := False;
                     end if;
                  end if;
               end loop;
               if Valid_Last_Slice then
                  Texts.Write
                    (Text   => Result,
                     Cursor => Cursor,
                     Item   => Pattern (Slice_Start .. Pattern'Last));
               end if;
            end return;
         end Return_Result;
      end Common_Case;
   end Variable_Message;

end Tiny.Texts;
