
-- Combines features for characters and strings handling, which do not
-- exist in Ada standard packages. Character and string test and convertion,
-- with a word/blank tokenizer. Note Is_Graphic for Character is redefined.

with Ada.Characters.Latin_1;

package Tiny.Texts is

   -- ========================================================================
   -- Table of Content
   -- ------------------------------------------------------------------------
   -- 1. Definitions
   -- 2. Tests
   -- 3. Convertions
   -- 4. Edition
   -- 5. Functions
   -- 6. Tokenizer
   -- 7. Variable substitutions
   -- ------------------------------------------------------------------------

   -- ========================================================================
   -- 1. Definitions
   -- ------------------------------------------------------------------------

   subtype Position_Type is Positive;
   -- Subtype of the String index type.

   -- The following named numbers get bound attributes as universal
   -- integer.

   Position_Type_First : constant := Position_Type'First;
   Position_Type_Last  : constant := Position_Type'Last;

   subtype String_Length_Type is Integer range
     (0) ..
     (Position_Type_Last - Position_Type_First + 1);
   -- The biggest length with a string, is the biggest length one
   -- can get between two included bounds of a string.

   type String_Access is access constant String;
   -- Useful every where an constrained type is required while a desired
   -- string entity cannot be constrained (most famously in arrays and
   -- records).

   type Strings_Type is array
     (Positive range <>) of
     access constant String;

   Blank : constant Character := Ada.Characters.Latin_1.Space;

   Empty_String : constant String := "";
   -- Whatever are the bounds, it is just required the length is zero.

   Wide_Replacement_Character : constant Wide_Character :=
     Wide_Character'Val (16#FFFD#);
   -- Replacement character as defined by the Unicode standard.

   ASCII_Replacement_Character : constant Character := '?';
   -- The Unicode standard defined replacement character, is outside of
   -- the ASCII range, a replacement-replacement-character is to be
   -- defined for ASCII/Standard String targets.

   -- ========================================================================
   -- 2. Tests
   -- ------------------------------------------------------------------------

   function Is_Graphic --!renomer en Is_Not_Blank
     (C : Character)
      return Boolean;
   -- Returns the same as Ada.Characters.Handling.Is_Graphic, except
   -- for spacs : this returns False if C is the space character.

   function Is_Blank
     (C : Character)
      return Boolean;
   --|Ensures: Is_Blank (C) = not Is_Not_Bkank (C)

   function Is_Graphic (S : String) return Boolean;
   --|Ensures:
   --   (for each I in S'Range, Is_Graphic (Long_Name (I))) <=>
   --   (Result = True)
   -- True means *all* characters are graphic.

   function Is_Blank (S : String) return Boolean;
   --|Ensures:
   --   (not (for each I in S'Range, Is_Graphic (Long_Name (I)))) <=>
   --   (Result = True)
   -- True means *at least one* character is not blank.

   -- ========================================================================
   -- 3. Convertions
   -- ------------------------------------------------------------------------

   function To_String (C : Character) return String;
   --|Ensures: Result'Length = 1
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result (Positive'First) = C

   function To_String
     (Source                : Wide_String;
      Replacement_Character : Character := ASCII_Replacement_Character)
      return String;
   --|Ensures: Result'Range = Source'Range
   -- Characters from source outside of the range of Character will be
   -- substituted the Replacement_Character.

   function To_Wide_String (Source : String) return Wide_String;
   --|Ensures: Result'Range = Source'Range

   -- ========================================================================
   -- 4. Edition
   -- ------------------------------------------------------------------------

   procedure Write
     (Text   : in out String;
      Cursor : in out Positive;
      Item   : in String);
   --|Requires: Text'Length > 0
   --|Requires: Cursor in Text'Range
   -- This implies Text cannot be an empty string, and this implication is
   -- left as part of the specification.
   --|Requires:
   --   (Item'Length > 0) =>
   --   ((Item'Length - 1) <= (Text'Last - Cursor))
   -- This means it is required the Item can be written to Text starting
   -- at Cursor, without any overflow.
   --|Ensures:
   --|   Cursor =
   --|   (Text'First + ((old Cursor + Item'Length) mod Text'Length))
   -- This ensures new Cursor is always in Text'Range and move in a consistent
   -- way. The other way to specify how Cursor is updated, would have been
   -- to stick the cursor at Text'Last as soon an Item is written up to
   -- the end of Text. Methametacally speaking, this is not aesthetic. Modulo
   -- arithmetic like behaviour it more clean, and may also be useful, as this
   -- reset the Cursor position, as soon as an Item is is written up to the
   -- end, thus, as soon as Text is probably fully written in all of its
   -- range.

   procedure Write
     (Text   : in out String;
      Cursor : in out Positive;
      Item   : in Character);
   --|Requires: Text'Length > 0
   --|Requires: Cursor in Text'Range
   -- This implies Text cannot be an empty string, and this implication is
   -- left as part of the specification.
   -- Note: as Cursor is required to be in Text'Range and the written Item
   -- is a character, that is, an item of length 1, there is not need
   -- for a require assertion like the one in the same function which gets
   -- a Item of type String wich ensure the Item can be written without any
   -- overflow. Indeed, if Cursor is in Text'Range, it is sufficient to
   -- ensure the Character Item can be written without overflow.
   -- This means it is required the Item can be written to Text starting
   -- at Cursor, without any overflow.
   --|Ensures: Cursor = Text'First + ((old Cursor + 1) mod Text'Length)
   -- See comment in the other Write function.

   -- ========================================================================
   -- 5. Functions
   -- ------------------------------------------------------------------------

   function Blank_String (Length : Natural) return String;
   --|Requires: Length < Natural'Last
   --|Ensures: Result'Length = Length
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result (Positive'First .. Length) = Blank

   function Each_Appended (Items : Strings_Type) return String;
   --|Requires: Sum (Items(each I)'Length) <= String_Length_Type'Last
   --|Ensures: (Items'Length = 0) -> Result'Length = 0
   --|Ensures: Result'First = Positive'First
   --|Ensures: Result'Length = Sum (Items(each I)'Length)

   -- ========================================================================
   -- 6. Tokenizer
   -- ------------------------------------------------------------------------

   type Token_Accessor_Type is abstract tagged private;

   type Token_Kinds is
     (A_Blank,
      A_Word);

   procedure Do_Access
     (Accessor : in out Token_Accessor_Type;
      Token    : in String;
      Kind     : in Token_Kinds)
   is abstract;
   --|Requires: Token'Length > 0

   subtype Token_Accessor_Class is Token_Accessor_Type'Class;

   procedure For_Each_Word_And_Blank
     (Source   : in String;
      Accessor : in out Token_Accessor_Class);
   --|Ensures: when Do_Access (Accessor, Token, Kind), Token'Length > 0
   --|Ensures:
   --|   when Do_Access[n+1] (Accessor, Token, Kind),
   --|   Kind[n] /= Kind[n+1]
   -- Do_Access will be invoked with alternatively A_Blank, A_Word. This
   -- is implied by the definition of what A_Blank and A_Word each stands for

   -- ========================================================================
   -- 7. Variable substitutions
   -- ------------------------------------------------------------------------
   function Variable_Message
     (Pattern : String;
      Value   : String)
      return String;
   -- Return Pattern with all occurences of "$1" substituted to Value.

   -- ========================================================================
   -- ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CROSS ***** DO NOT CR
   -- ========================================================================

private

   type Token_Accessor_Type is abstract
     tagged null record;

end Tiny.Texts;
